# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import io
import json
from pathlib import Path
from abc import ABC, abstractmethod
from typing import Any, Union, Iterable, Optional, Dict

EmbeddingVector = Iterable[float]


class Vectorizer(ABC):
    """
    Vectorizer turns texts into embedding vectors.
    """

    def __init__(self, config: Dict[str, Any]):
        self._config = config
        self._vector_dimensions = None

    @classmethod
    def from_config(cls, config: Union[str, Path, Dict[str, Any]]) -> "Vectorizer":
        """
        Create vectorizer from `config`.

        If `config` is a string or path, it will be loaded as a dictionary depending
        on its file extension. Currently, the following formats are supported:

        * .json: JSON
        * .json5: JSON with comments support
        * .yaml: YAML

        :param config: vectorizer config
        :type config: str, Path or Dict[str, Any]
        :return: vectorizer instance
        :rtype: Vectorizer
        """
        from kag.common.utils import dynamic_import_class

        if isinstance(config, (str, Path)):
            config_path = config
            if not isinstance(config_path, Path):
                config_path = Path(config_path)
            if config_path.name.endswith(".yaml"):
                import yaml

                with io.open(config_path, "r", encoding="utf-8") as fin:
                    config = yaml.safe_load(fin)
            elif config_path.name.endswith(".json5"):
                import json5

                with io.open(config_path, "r", encoding="utf-8") as fin:
                    config = json5.load(fin)
            elif config_path.name.endswith(".json"):
                with io.open(config_path, "r", encoding="utf-8") as fin:
                    config = json.load(fin)
            else:
                message = "only .json, .json5 and .yaml are supported currently; "
                message += "can not load vectorizer config from %r" % str(config_path)
                raise RuntimeError(message)
        elif isinstance(config, dict):
            pass
        else:
            message = "only str, Path and dict are supported; "
            message += "invalid vectorizer config: %r" % (config,)
            raise RuntimeError(message)

        class_name = config.get("vectorizer")
        if class_name is None:
            message = "vectorizer class name is not specified"
            raise RuntimeError(message)
        vectorizer_class = dynamic_import_class(class_name, "vectorizer")
        if not issubclass(vectorizer_class, Vectorizer):
            message = "class %r is not a vectorizer class" % (class_name,)
            raise RuntimeError(message)
        vectorizer = vectorizer_class._from_config(config)
        return vectorizer

    @classmethod
    @abstractmethod
    def _from_config(cls, config: Dict[str, Any]) -> "Vectorizer":
        """
        Create vectorizer from `config`. This method is supposed to be implemented
        by derived classes.

        :param config: vectorizer config
        :type config: Dict[str, Any]
        :return: vectorizer instance
        :rtype: Vectorizer
        """
        message = "abstract method _from_config is not implemented"
        raise NotImplementedError(message)

    def _get_vector_dimensions(self, config: Dict[str, Any]) -> Optional[int]:
        """
        Get embedding vector dimensions from `config`.

        * If vector dimensions is not specified in `config`, return None.

        * If vector dimensions is specified in `config` but not a positive integer,
          raise an exception.

        :param config: vectorizer config
        :type config: Dict[str, Any]
        :return: embedding vector dimensions or None
        :rtype: Optional[int]
        """
        value = config.get("vector_dimensions")
        if value is None:
            return None
        if isinstance(value, str):
            try:
                value = int(value)
            except ValueError as ex:
                message = "vector_dimensions must be integer; "
                message += "%r is invalid" % (value,)
                raise RuntimeError(message) from ex
        if not isinstance(value, int) or value <= 0:
            message = "vector_dimensions must be positive-integer; "
            message += "%r is invalid" % (value,)
            raise RuntimeError(message)
        return value

    @property
    def vector_dimensions(self):
        """
        Dimension of generated embedding vectors.
        """
        if self._vector_dimensions is not None:
            return self._vector_dimensions
        try:
            example_input = "This is a test."
            example_vector = self.vectorize(example_input)
        except Exception as ex:
            message = "the embedding service is not available"
            raise RuntimeError(message) from ex
        value = self._get_vector_dimensions(self._config)
        if value is not None and value != len(example_vector):
            message = "invalid 'vector_dimensions', specified %d; " % value
            message += "but the actual generated embedding vector is of %d dimensions" % len(example_vector)
            raise RuntimeError(message)
        self._vector_dimensions = len(example_vector)
        return self._vector_dimensions

    @abstractmethod
    def vectorize(self, texts: Union[str, Iterable[str]]) -> Union[EmbeddingVector, Iterable[EmbeddingVector]]:
        """
        Vectorize a text string into an embedding vector or multiple text strings into
        multiple embedding vectors.

        :param texts: texts to vectorize
        :type texts: str or Iterable[str]
        :return: embedding vectors of the texts
        :rtype: EmbeddingVector or Iterable[EmbeddingVector]
        """
        message = "abstract method vectorize is not implemented"
        raise NotImplementedError(message)



import logging

from kag.common.base.prompt_op import PromptOp
from kag.solver.common.base import KagBaseModule

logger = logging.getLogger(__name__)


class LFGenerator(KagBaseModule):
    """
    Generator class that selects different prompts based on the scenario to produce answers.
    This class can be extended to implement custom generation strategies.
    """

    def __init__(self,**kwargs):
        super().__init__(**kwargs)
        self.solve_question_prompt = PromptOp.load(self.biz_scene, "solve_question")(
            language=self.language
        )

        self.solve_question_without_docs_prompt = PromptOp.load(self.biz_scene, "solve_question_without_docs")(
            language=self.language
        )

        self.solve_question_without_spo_prompt = PromptOp.load(self.biz_scene, "solve_question_without_spo")(
            language=self.language
        )

    def generate_sub_answer(self, question: str, knowledge_graph: [], docs: [], history=[]):
        """
        Generates a sub-answer based on the given question, knowledge graph, documents, and history.

        Parameters:
        question (str): The main question to answer.
        knowledge_graph (list): A list of knowledge graph data.
        docs (list): A list of documents related to the question.
        history (list, optional): A list of previous query-answer pairs. Defaults to an empty list.

        Returns:
        str: The generated sub-answer.
        """
        history_qa = [f"query{i}: {item['sub_query']}\nanswer{i}: {item['sub_answer']}" for i, item in
                      enumerate(history)]
        if knowledge_graph:
            if len(docs) > 0:
                prompt = self.solve_question_prompt
                params = {
                    'question': question,
                    'knowledge_graph': str(knowledge_graph),
                    'docs': str(docs),
                    'history': '\n'.join(history_qa)
                }
            else:
                prompt = self.solve_question_without_docs_prompt
                params = {
                    'question': question,
                    'knowledge_graph': str(knowledge_graph),
                    'history': '\n'.join(history_qa)
                }
        else:
            prompt = self.solve_question_without_spo_prompt
            params = {
                'question': question,
                'docs': str(docs),
                'history': '\n'.join(history_qa)
            }
        llm_output = self.llm_module.invoke(params, prompt, with_json_parse=False, with_except=True)
        logger.debug(f"sub_question:{question}\n sub_answer:{llm_output} prompt:\n{prompt}")
        if llm_output:
            return llm_output
        return "I don't know"



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import json
from collections import defaultdict
from typing import List, Type, Dict

from kag.interface.builder.mapping_abc import MappingABC
from kag.builder.model.sub_graph import SubGraph
from knext.common.base.runnable import Input, Output
from knext.schema.client import OTHER_TYPE


class SPOMapping(MappingABC):

    def __init__(self):
        super().__init__()
        self.s_type_col = None
        self.s_id_col = None
        self.p_type_col = None
        self.o_type_col = None
        self.o_id_col = None
        self.sub_property_mapping = defaultdict(list)
        self.sub_property_col = None

    @property
    def input_types(self) -> Type[Input]:
        return Dict[str, str]

    @property
    def output_types(self) -> Type[Output]:
        return SubGraph

    def add_field_mappings(self, s_id_col: str, p_type_col: str, o_id_col: str, s_type_col: str = None, o_type_col: str = None):
        self.s_type_col = s_type_col
        self.s_id_col = s_id_col
        self.p_type_col = p_type_col
        self.o_type_col = o_type_col
        self.o_id_col = o_id_col
        return self

    def add_sub_property_mapping(self, source_name: str, target_name: str = None):
        """
        Adds a field mapping from source data to a property of the subject type.

        Args:
            source_name (str): The source field to be mapped.
            target_name (str): The target field to map the source field to.

        Returns:
            self
        """
        if self.sub_property_col:
            raise ValueError("Fail to add sub property mapping.")
        if not target_name:
            self.sub_property_col = source_name
        else:
            self.sub_property_mapping[target_name].append(source_name)
        return self

    def assemble_sub_graph(self, record: Dict[str, str]):
        """
        Assembles a subgraph from the provided record.

        Args:
            record (Dict[str, str]): The record containing the data to assemble into a subgraph.

        Returns:
            SubGraph: The assembled subgraph.
        """
        sub_graph = SubGraph(nodes=[], edges=[])
        s_type = record.get(self.s_type_col) or OTHER_TYPE
        s_id = record.get(self.s_id_col) or ""
        p = record.get(self.p_type_col) or ""
        o_type = record.get(self.o_type_col) or OTHER_TYPE
        o_id = record.get(self.o_id_col) or ""
        sub_graph.add_node(id=s_id, name=s_id, label=s_type)
        sub_graph.add_node(id=o_id, name=o_id, label=o_type)
        sub_properties = {}
        if self.sub_property_col:
            sub_properties = json.loads(record.get(self.sub_property_col, '{}'))
            sub_properties = {k: str(v) for k, v in sub_properties.items()}
        else:
            for target_name, source_names in self.sub_property_mapping.items():
                for source_name in source_names:
                    value = record.get(source_name)
                    sub_properties[target_name] = value
        sub_graph.add_edge(s_id=s_id, s_label=s_type, p=p, o_id=o_id, o_label=o_type, properties=sub_properties)
        return sub_graph

    def invoke(self, input: Input, **kwargs) -> List[Output]:
        """
        Invokes the mapping process on the given input and returns the resulting sub-graphs.

        Args:
            input (Input): The input data to be processed.
            **kwargs: Additional keyword arguments.

        Returns:
            List[Output]: A list of resulting sub-graphs.
        """
        record: Dict[str, str] = input
        sub_graph = self.assemble_sub_graph(record)
        return [sub_graph]



import logging
import re

from kag.solver.logic.core_modules.common.base_model import SPOBase, SPOEntity, SPORelation, Identifer, \
    TypeInfo, LogicNode
from kag.solver.logic.core_modules.common.schema_utils import SchemaUtils

logger = logging.getLogger(__name__)\


# get_spg(s, p, o)
class GetSPONode(LogicNode):
    def __init__(self, operator, args):
        super().__init__(operator, args)
        self.s: SPOBase = args.get('s', None)
        self.p: SPOBase = args.get('p', None)
        self.o: SPOBase = args.get('o', None)
        self.sub_query = args.get("sub_query", None)
        self.query = args.get("query", None)

    def to_dsl(self):
        raise NotImplementedError("Subclasses should implement this method.")

    def to_std(self, args):
        for key, value in args.items():
            self.args[key] = value
        self.s = args.get('s', self.s)
        self.p = args.get('p', self.p)
        self.o = args.get('o', self.o)
        self.sub_query = args.get('sub_query', self.sub_query)

    @staticmethod
    def parse_node(input_str):
        equality_list = re.findall(r'([\w.]+=[^=]+)(,|，|$)', input_str)
        if len(equality_list) < 3:
            raise RuntimeError(f"parse {input_str} error not found s,p,o")
        spo_params = [e[0] for e in equality_list[:3]]
        get_spo_node = GetSPONode.parse_node_spo(spo_params)
        if len(equality_list) > 3:
            value_params = [e[0] for e in equality_list[3:]]
            GetSPONode.parse_node_value(get_spo_node, value_params)
        return get_spo_node

    @staticmethod
    def parse_node_spo(spo_params):
        s = None
        p = None
        o = None
        for spo_param in spo_params:
            key, param = spo_param.split('=')
            if key == "s":
                s = SPOEntity.parse_logic_form(param)
            elif key == "o":
                o = SPOEntity.parse_logic_form(param)
            elif key == "p":
                p = SPORelation.parse_logic_form(param)
        if s is None:
            raise RuntimeError(f"parse {str(spo_params)} error not found s")
        if p is None:
            raise RuntimeError(f"parse {str(spo_params)} error not found p")
        if o is None:
            raise RuntimeError(f"parse {str(spo_params)} error not found o")
        return GetSPONode("get_spo", {
            "s": s,
            "p": p,
            "o": o
        })

    @staticmethod
    def parse_node_value(get_spo_node_op, value_params):
        for value_param in value_params:
            # a.value=123,b.brand=345
            value_pair = re.findall(r'(?:[,\s]*(\w+)\.(\w+)=([^,，]+))', value_param)
            for key, property, value in value_pair:
                node = None
                if key == "s":
                    node = get_spo_node_op.s
                elif key == "p":
                    node = get_spo_node_op.p
                elif key == "o":
                    node = get_spo_node_op.o
                node.value_list.append([str(property), value])


def binary_expr_parse(input_str):
    pattern = re.compile(r'(\w+)=((?:(?!\w+=).)*)')
    matches = pattern.finditer(input_str)
    left_expr = None
    right_expr = None
    op = None
    for match in matches:
        key = match.group(1).strip()
        value = match.group(2).strip().rstrip(',')
        value = value.rstrip('，')
        if key == "left_expr":
            if "," in value:
                left_expr_list = list(set([Identifer(v) for v in value.split(",")]))
            elif "，" in value:
                left_expr_list = list(set([Identifer(v) for v in value.split("，")]))
            else:
                left_expr_list = [Identifer(value)]
            if len(left_expr_list) == 1:
                left_expr = left_expr_list[0]
            else:
                left_expr = left_expr_list
        elif key == "right_expr":
            if value != '':
                right_expr = value
        elif key == "op":
            op = value
    if left_expr is None:
        raise RuntimeError(f"parse {input_str} error not found left_expr")

    if op is None:
        raise RuntimeError(f"parse {input_str} error not found op")
    return {
        "left_expr": left_expr,
        "right_expr": right_expr,
        "op": op
    }


# filter(left_expr=alias, right_expr=other_alias or const_data, op=equal|lt|gt|le|ge|in|contains|and|or|not)
class FilterNode(LogicNode):
    def __init__(self, operator, args):
        super().__init__(operator, args)
        self.left_expr = args.get('left_expr', None)
        self.right_expr = args.get('right_expr', None)
        self.op = args.get('op', None)
        self.OP = 'equal|lt|gt|le|ge|in|contains|and|or|not'.split('|')

    def to_dsl(self):
        raise NotImplementedError("Subclasses should implement this method.")

    def to_std(self, args):
        for key, value in args.items():
            self.args[key] = value
        self.left_expr = args.get('left_expr', self.left_expr)
        self.right_expr = args.get('right_expr', self.right_expr)
        self.op = args.get('op', self.op)

    @staticmethod
    def parse_node(input_str):
        args = binary_expr_parse(input_str)
        return FilterNode("filter", args)


# count(alias)->count_alias
class CountNode(LogicNode):
    def __init__(self, operator, args):
        super(CountNode, self).__init__(operator, args)
        self.alias_name = args.get("alias_name", None)
        self.set = args.get("set", None)

    def to_dsl(self):
        raise NotImplementedError("Subclasses should implement this method.")

    @staticmethod
    def parse_node(input_str, output_name):
        args = {'alias_name': output_name, 'set': input_str}
        return CountNode("count", args)


# sum(alias)->sum_alias
class SumNode(LogicNode):
    def __init__(self, operator, args):
        super(SumNode, self).__init__(operator, args)
        self.alias_name = args.get("alias_name", None)
        self.set = args.get("set", None)

    def to_dsl(self):
        raise NotImplementedError("Subclasses should implement this method.")

    @staticmethod
    def parse_node(input_str):
        # count_alias=count(alias)
        match = re.match(r'(\w+)[\(\（](.*)[\)\）](->)?(.*)?', input_str)
        if not match:
            raise RuntimeError(f"parse logic form error {input_str}")
        # print('match:',match.groups())
        if len(match.groups()) == 4:
            operator, params, _, alias_name = match.groups()
        else:
            operator, params = match.groups()
            alias_name = 'sum1'
        params = params.replace('，', ',').split(',')
        args = {'alias_name': alias_name, 'set': params}
        return SumNode("sum", args)


class SortNode(LogicNode):
    def __init__(self, operator, args):
        super().__init__(operator, args)
        self.alias_name = args.get("alias_name", None)
        self.set = args.get("set", None)
        self.orderby = args.get("orderby", None)
        self.direction = args.get("direction", None)
        self.limit = args.get("limit", None)

    def to_dsl(self):
        raise NotImplementedError("Subclasses should implement this method.")

    def __str__(self):
        return f"sort：{self.set} {self.orderby} {self.direction} top{self.limit}"

    def get_set(self):
        if isinstance(self.set, list):
            return set(self.set)
        return {self.set}

    @staticmethod
    def parse_node(input_str):
        equality_list = re.findall(r'([\w.]+=[^=]+)(,|，|$)', input_str)
        if len(equality_list) < 4:
            raise RuntimeError(f"parse {input_str} error not found set,orderby,direction,limit")
        params = [e[0] for e in equality_list[:4]]
        params_dict = {}
        for param in params:
            key, value = param.split('=')
            params_dict[key] = value
        return SortNode("sort", params_dict)


class CompareNode(LogicNode):
    def __init__(self, operator, args):
        super().__init__(operator, args)
        self.alias_name = args.get("alias_name", None)
        self.set = args.get("set", None)
        self.op = args.get("op", None)

    def to_dsl(self):
        raise NotImplementedError("Subclasses should implement this method.")

    def __str__(self):
        return f"compare：{self.set} {self.op} "

    def get_set(self):
        if isinstance(self.set, list):
            return set(self.set)
        return {self.set}

    @staticmethod
    def parse_node(input_str):
        equality_list = re.findall(r'([\w.]+=[^=]+)(,|，|$)', input_str)
        if len(equality_list) < 2:
            raise RuntimeError(f"parse {input_str} error not found set,orderby,direction,limit")
        params = [e[0] for e in equality_list[:2]]
        params_dict = {}
        for param in params:
            key, value = param.split('=')
            if key == 'set':
                value = value.strip().replace('，', ',').replace(' ', '').strip('[').strip(']').split(',')
            params_dict[key] = value
        return CompareNode("compare", params_dict)


class DeduceNode(LogicNode):
    def __init__(self, operator, args):
        super().__init__(operator, args)
        self.deduce_ops = args.get("deduce_ops", [])

    def __str__(self):
        return f"deduce(op={','.join(self.deduce_ops)})"

    @staticmethod
    def parse_node(input_str):
        ops = input_str.replace("op=", "")
        input_ops = ops.split(",")
        return DeduceNode("deduce", {
            "deduce_ops": input_ops
        })


# verity(left_expr=alias, right_expr=other_alias or const_data, op=equal|gt|lt|ge|le|in|contains)
class VerifyNode(LogicNode):
    def __init__(self, operator, args):
        super().__init__(operator, args)
        self.left_expr = args.get('left_expr', None)
        self.right_expr = args.get('right_expr', None)
        self.op = args.get('op', None)
        self.OP = {'等于': 'equal', '大于': 'gt', '小于': 'lt', '大于等于': 'ge', '小于等于': 'le', '属于': 'in', '包含': 'contains'}

    def to_dsl(self):
        raise NotImplementedError("Subclasses should implement this method.")

    def __str__(self):
        return f"条件判定：{self.left_expr} {self.op} {self.right_expr}"

    def get_left_expr_set(self):
        if isinstance(self.left_expr, list):
            return set(self.left_expr)
        return {self.left_expr}

    def get_left_expr_name(self):
        if isinstance(self.left_expr, list):
            return ",".join([str(e) for e in self.left_expr])
        return self.left_expr

    def to_std(self, args):
        for key, value in args.items():
            self.args[key] = value
        self.left_expr = args.get('left_expr', self.left_expr)
        self.right_expr = args.get('right_expr', self.right_expr)
        self.op = args.get('op', self.op)
        if self.op in self.OP.values():
            self.op = self.OP[self.op]

    @staticmethod
    def parse_node(input_str):
        if "verify" in input_str:
            match = re.match(r'(\w+)[\(\（](.*)[\)\）](->)?(.*)?', input_str)
            if not match:
                raise RuntimeError(f"parse logic form error {input_str}")
            # print('match:',match.groups())
            if len(match.groups()) == 4:
                operator, input_str, _, output_name = match.groups()
            else:
                operator, input_str = match.groups()
        args = binary_expr_parse(input_str)
        return VerifyNode("verify", args)


class ExtractorNode(LogicNode):
    def __int__(self, operator, args):
        super(ExtractorNode, self).__init__(operator, args)
        self.alias_set = args.get("alias_set")

    def to_dsl(self):
        raise NotImplementedError("Subclasses should implement this method.")

    @staticmethod
    def parse_node(input_str):
        params = set(input_str.split(","))
        alias_set = [Identifer(p) for p in params]
        ex_node = ExtractorNode("extractor", {
            "alias_set": alias_set
        })
        ex_node.alias_set = alias_set
        return ex_node


# get(alias_name)
class GetNode(LogicNode):
    def __init__(self, operator, args):
        super(GetNode, self).__init__(operator, args)
        self.alias_name = args.get("alias_name")
        self.alias_name_set: list = args.get("alias_name_set")
        self.s = args.get("s", None)
        self.s_alias_map: dict = args.get("s_alias_map", None)

    def to_dsl(self):
        raise NotImplementedError("Subclasses should implement this method.")

    @staticmethod
    def parse_node(input_str):
        input_args = input_str.split(",")
        return GetNode("get", {
            "alias_name": Identifer(input_args[0]),
            "alias_name_set": [Identifer(e) for e in input_args]
        })


# search_s()
class SearchNode(LogicNode):
    def __init__(self, operator, args):
        super().__init__(operator, args)
        self.s = SPOEntity(None, None, args['type'], None, args['alias'], False)
        self.s.value_list = args['conditions']

    @staticmethod
    def parse_node(input_str):
        pattern = re.compile(r'[,\s]*s=(\w+):([^,\s]+),(.*)')
        matches = pattern.match(input_str)
        args = dict()
        args["alias"] = matches.group(1)
        args["type"] = matches.group(2)
        if len(matches.groups()) > 2:
            search_condition = dict()
            s_condition = matches.group(3)

            condition_pattern = re.compile(r'(?:[,\s]*(\w+)\.(\w+)=([^,，]+))')
            condition_list = condition_pattern.findall(s_condition)
            for condition in condition_list:
                s_property = condition[1]
                s_value = condition[2]
                s_value = SearchNode.check_value_is_reference(s_value)
                search_condition[s_property] = s_value
            args['conditions'] = search_condition

        return SearchNode('search_s', args)

    @staticmethod
    def check_value_is_reference(value_str):
        if '.' in value_str:
            return value_str.split('.')
        return value_str


class ParseLogicForm:
    def __init__(self, schema: SchemaUtils, schema_retrieval):
        self.schema = schema
        self.schema_retrieval = schema_retrieval

    def std_parse_kg_node(self, entity: SPOBase, parsed_entity_set):
        alias_name = entity.alias_name
        if alias_name in parsed_entity_set.keys():
            exist_node = parsed_entity_set[alias_name]
            exist_node.value_list.extend(entity.value_list)
            return parsed_entity_set[alias_name]

        zh_types = entity.get_entity_type_zh_set()
        std_entity_type_set = []
        if isinstance(entity, SPOEntity):
            for entity_type in zh_types:
                type_info = self.get_node_type_info(entity_type)
                if type_info.entity_type is None and self.schema is not None:
                    entity.is_attribute = True
                std_entity_type_set.append(type_info)
        elif isinstance(entity, SPORelation):
            s_type_zh = entity.s.get_entity_first_type_or_en()
            o_type_zh = entity.o.get_entity_first_type_or_en()
            s_type_en = entity.s.get_entity_first_type_or_zh()
            o_type_en = entity.o.get_entity_first_type_or_zh()
            for entity_type in zh_types:
                type_info = TypeInfo()
                type_info.entity_type_zh = entity_type
                if self.schema is not None:
                    if o_type_zh == "Entity":
                        sp_index = (s_type_zh, entity_type)
                        if sp_index in self.schema.sp_o:
                            o_candis_set = self.schema.sp_o[sp_index]
                            for candis in o_candis_set:
                                spo_zh = f"{s_type_zh}_{entity_type}_{candis}"
                                type_info.entity_type = self.schema.get_spo_with_p(self.schema.spo_zh_en[spo_zh])
                                break

                    if not type_info.entity_type and s_type_zh == "Entity":
                        op_index = (o_type_zh, entity_type)
                        if op_index in self.schema.op_s:
                            s_candis_set = self.schema.op_s[op_index]
                            for candis in s_candis_set:
                                spo_zh = f"{candis}_{entity_type}_{o_type_zh}"
                                type_info.entity_type = self.schema.get_spo_with_p(self.schema.spo_zh_en[spo_zh])
                                break

                    if not type_info.entity_type and o_type_zh != "Entity" and s_type_zh != "Entity":
                        so_index = (s_type_zh, o_type_zh)
                        if so_index not in self.schema.so_p:
                            so_index = (o_type_zh, s_type_zh)
                        candis_set = self.schema.so_p[so_index]
                        for p_candis in candis_set:
                            if p_candis == entity_type:
                                spo_zh = f"{s_type_zh}_{p_candis}_{o_type_zh}"
                                type_info.entity_type = self.schema.get_spo_with_p(self.schema.spo_zh_en[spo_zh])

                    if not type_info.entity_type:
                        # maybe a property
                        s_attr_zh_en = self.schema.attr_zh_en_by_label.get(s_type_en, [])
                        if s_attr_zh_en and entity_type in s_attr_zh_en:
                            type_info.entity_type = s_attr_zh_en[entity_type]
                        if not type_info.entity_type:
                            o_attr_zh_en = self.schema.attr_zh_en_by_label.get(o_type_en, [])
                            if o_attr_zh_en and entity_type in o_attr_zh_en:
                                type_info.entity_type = o_attr_zh_en[entity_type]
                std_entity_type_set.append(type_info)

        entity.type_set = std_entity_type_set
        parsed_entity_set[alias_name] = entity

        return entity

    def std_parse_node(self, entity: SPOEntity, parsed_entity_set):
        alias_name = entity.alias_name
        if alias_name in parsed_entity_set.keys():
            exist_node = parsed_entity_set[alias_name]
            exist_node.value_list.extend(entity.value_list)
            return parsed_entity_set[alias_name]

        zh_types = entity.get_entity_type_zh_set()
        std_entity_type_set = []
        for entity_type in zh_types:
            type_info = self.get_node_type_info(entity_type)
            if type_info.entity_type is None and self.schema is not None:
                entity.is_attribute = True
            std_entity_type_set.append(type_info)
        entity.type_set = std_entity_type_set
        parsed_entity_set[alias_name] = entity
        return entity

    def std_parse_edge(self, edge: SPORelation, parsed_entity_set):
        alias_name = edge.alias_name
        if alias_name in parsed_entity_set.keys():
            return parsed_entity_set[alias_name]
        zh_types = edge.get_entity_type_zh_set()
        std_edge_type_set = []
        for entity_type in zh_types:
            type_info = self.get_edge_type_info(entity_type)
            if type_info.entity_type is None and self.schema is not None:
                edge.is_attribute = True
            std_edge_type_set.append(type_info)
        edge.type_set = std_edge_type_set

        parsed_entity_set[alias_name] = edge
        return edge

    def parse_logic_form(self, input_str: str, parsed_entity_set={}, sub_query=None, query=None):
        match = re.match(r'(\w+)[\(\（](.*)[\)\）](->)?(.*)?', input_str.strip())
        if not match:
            raise RuntimeError(f"parse logic form error {input_str}")
        if len(match.groups()) == 4:
            operator, args_str, _, output_name = match.groups()
        else:
            operator, args_str = match.groups()
            output_name = None
        low_operator = operator.lower()
        if low_operator == "get":
            node: GetNode = GetNode.parse_node(args_str)
            if node.alias_name in parsed_entity_set.keys():
                s = parsed_entity_set[node.alias_name]
                node.s = s
        elif low_operator in ["get_spo", "retrieval"]:
            node: GetSPONode = GetSPONode.parse_node(args_str)
            s_node = self.std_parse_kg_node(node.s, parsed_entity_set)
            o_node = self.std_parse_kg_node(node.o, parsed_entity_set)
            node.p.s = s_node
            node.p.o = o_node
            p_node = self.std_parse_kg_node(node.p, parsed_entity_set)
            node.to_std({
                "s": s_node,
                "p": p_node,
                "o": o_node,
                "sub_query": sub_query,
            })
        elif low_operator in ["filter"]:
            node: FilterNode = FilterNode.parse_node(args_str)
        elif low_operator in ["deduce"]:
            node: DeduceNode = DeduceNode.parse_node(args_str)
        elif low_operator in ["verify"]:
            node: VerifyNode = VerifyNode.parse_node(args_str)
        elif low_operator in ["count"]:
            node: CountNode = CountNode.parse_node(args_str, output_name)
        elif low_operator in ["sum"]:
            node: SumNode = SumNode.parse_node(args_str)
        elif low_operator in ["sort"]:
            node: SortNode = SortNode.parse_node(args_str)
        elif low_operator in ["compare"]:
            node: SortNode = CompareNode.parse_node(args_str)
        elif low_operator in ["extractor"]:
            node: ExtractorNode = ExtractorNode.parse_node(args_str)
        elif low_operator in ['search_s']:
            node: SearchNode = SearchNode.parse_node(args_str)
            self.std_parse_node(node.s, parsed_entity_set)
        else:
            raise NotImplementedError(f"not impl {input_str}")

        node.to_std({
            "sub_query": sub_query
        })

        return node

    def parse_logic_form_set(self, input_str_set: list, sub_querys: list, question: str):
        parsed_cached_map = {}
        parsed_node = []
        for i, input_str in enumerate(input_str_set):
            if sub_querys and i < len(sub_querys):
                sub_query = sub_querys[i]
            else:
                sub_query = None
            try:
                logic_node = self.parse_logic_form(input_str, parsed_cached_map, sub_query=sub_query, query=question)
                parsed_node.append(logic_node)
            except Exception as e:
                logger.warning(f"parse node {input_str} error", exc_info=True)
        return parsed_node

    def std_node_type_name(self, type_name):
        if self.schema_retrieval is None:
            return type_name
        search_entity_labels = self.schema_retrieval.retrieval_entity(SPOEntity(entity_name=type_name))
        if len(search_entity_labels) > 0:
            return search_entity_labels[0].name
        return type_name

    def get_edge_type_en_by_name(self, type_name):
        if self.schema is None:
            return None
        if type_name in self.schema.edge_en_zh.keys():
            return type_name
        return self.schema.edge_zh_en.get(type_name, None)

    def get_node_type_en_by_name(self, type_name):
        if self.schema is None:
            return type_name
        if type_name in self.schema.node_en_zh.keys():
            return type_name
        return self.schema.node_zh_en.get(type_name, None)

    def get_node_type_zh_by_name(self, type_name):
        if self.schema is None:
            return type_name
        if type_name in self.schema.node_zh_en.keys():
            return type_name
        return self.schema.node_en_zh.get(type_name, None)

    def get_node_type_info(self, type_name):
        zh = self.get_node_type_zh_by_name(type_name)
        en = self.get_node_type_en_by_name(type_name)
        if zh == en:
            en = self.std_node_type_name(type_name)
            zh = en
        type_info = TypeInfo()
        type_info.entity_type = en
        type_info.entity_type_zh = zh
        if type_info.entity_type_zh is None:
            type_info.entity_type_zh = type_name
        return type_info

    def get_edge_type_info(self, type_name):
        # Edge is not standardized currently
        type_info = TypeInfo()
        type_info.entity_type = self.get_edge_type_en_by_name(type_name)
        type_info.entity_type_zh = type_name
        return type_info



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import os
from abc import ABC
from typing import List, Dict
import logging

from knext.common.base.component import Component
from knext.common.base.runnable import Input, Output
from knext.project.client import ProjectClient
from kag.common.llm.client import LLMClient


class BuilderComponent(Component, ABC):
    """
    Abstract base class for all builder component.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.project_id = kwargs.get("project_id",None) or os.getenv("KAG_PROJECT_ID")
        self.config = ProjectClient().get_config(self.project_id)


    def _init_llm(self) -> LLMClient:
        """
        Initializes the Large Language Model (LLM) client.

        This method retrieves the LLM configuration from environment variables and the project ID.
        It then fetches the project configuration using the project ID and updates the LLM configuration
        with any additional settings from the project. Finally, it creates and initializes the LLM client
        using the updated configuration.

        Args:
            None

        Returns:
            LLMClient
        """
        llm_config = eval(os.getenv("KAG_LLM", "{}"))
        project_id = self.project_id or os.getenv("KAG_PROJECT_ID")
        if project_id:
            try:
                config = ProjectClient().get_config(project_id)
                llm_config.update(config.get("llm", {}))
            except:
                logging.warning(
                    f"Failed to get project config for project id: {project_id}"
                )
        llm = LLMClient.from_config(llm_config)
        return llm

    @property
    def type(self):
        """
        Get the type label of the object.

        Returns:
            str: The type label of the object, fixed as "BUILDER".
        """
        return "BUILDER"

    def batch(self, inputs: List[Input], **kwargs) -> List[Output]:
        results = []
        for input in inputs:
            results.extend(self.invoke(input, **kwargs))
        return results

    def _handle(self, input: Dict) -> List[Dict]:
        _input = self.input_types.from_dict(input) if isinstance(input, dict) else input
        _output = self.invoke(_input)
        return [_o.to_dict() for _o in _output if _o]



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


import numpy as np
from typing import List


def rrf_score(length, r: int = 1):
    return np.array([1 / (r + i) for i in range(length)])


class Reranker:
    """
    This class provides a framework for a reranker,
    which is intended to re-rank the matches between queries and document passages.
    """

    def __init__(self):
        """
        Constructor for initializing the reranker class.
        Currently, there are no specific initialization parameters or operations.
        """
        pass

    def rerank(self, queries: List[str], passages: List[str]):
        """
        Function to re-rank queries and document passages,
        aiming to reorder the input query and passage pairs according to a certain strategy.

        Parameters:
        queries (List[str]): A list of strings containing queries that need to be re-ranked.
        passages (List[str]): A list of strings containing document passages that need to be re-ranked.

        The function is currently not implemented and raises an exception to indicate this.
        """
        raise NotImplementedError("rerank not implemented yet.")


import json
import logging
import re

from kag.solver.logic.core_modules.common.schema_utils import SchemaUtils

logger = logging.getLogger()


def find_and_extra_prop_objects(text):
    """
    Extracts and parses serialized objects from the given text.

    This function searches for all strings in the provided text that match the serialized format,
    then parses each found serialized string to extract object properties,
    and finally stores the parsed objects as dictionaries in a list which is returned.

    Parameters:
    text (str): The text containing serialized objects.

    Returns:
    list: A list of dictionaries representing the parsed objects.
    """

    pattern = re.compile(r'\001(.*?)\003')

    matches = pattern.findall(text)

    objects = []

    for match in matches:
        attributes = match.split('\002')
        if len(attributes) != 3:
            logger.info(f"find_and_extra_prop_objects attribute not match {match}")
            continue

        objects.append({
            "id": attributes[1],
            "name": attributes[0],
            "type": attributes[2]
        })

    return objects


class Prop:
    def __init__(self):
        self.origin_prop_map = {}
        self.extend_prop_map = {}
        self.linked_prop_map = {}

    def get_properties_map_list_value(self):
        result = {}
        for k in self.origin_prop_map.keys():
            result[k] = [self.origin_prop_map[k]]
        for k in self.extend_prop_map.keys():
            result[k] = [self.extend_prop_map[k]]
        return result

    @staticmethod
    def _get_ext_json_prop(schema: SchemaUtils):
        if schema is None:
            return []
        return schema.get_ext_json_prop()

    @staticmethod
    def _get_attr_en_zh_by_label(label_name, schema: SchemaUtils):
        if schema is None:
            return {}
        return schema.get_attr_en_zh_by_label(label_name)

    @staticmethod
    def from_dict(json_dict: dict, label_name: str, schema: SchemaUtils):
        prop = Prop()
        ext_attrs = Prop._get_ext_json_prop(schema)
        attr_en_zh = Prop._get_attr_en_zh_by_label(label_name, schema)
        black_attr = ["biz_node_id", "gdb_timestamp"]
        for k in json_dict.keys():
            if json_dict[k] == '' or k in black_attr:
                continue
            if k.startswith("_") or k in ext_attrs:
                continue
            if k in attr_en_zh.keys():
                key = attr_en_zh[k]
            else:
                key = k
            prop.origin_prop_map[key] = json_dict[k]

        for ext_attr in ext_attrs:
            if ext_attr not in json_dict.keys():
                continue
            try:
                basic_info = json.loads(json_dict[ext_attr])
                for k in basic_info.keys():
                    v = basic_info[k]
                    link_res = find_and_extra_prop_objects(v)
                    prop.linked_prop_map[k] = link_res
                prop.extend_prop_map = basic_info
            except Exception as e:
                logger.warning(f"parse basic info failed reasone: {json_dict[ext_attr]}", exc_info=True)
        return prop

    def to_json(self):
        return {
            "origin_prop_map": self.origin_prop_map,
            "extend_prop_map": self.extend_prop_map
        }

    def get_prop_value(self, p):
        if p in self.origin_prop_map.keys():
            return self.origin_prop_map[p]
        if p in self.extend_prop_map.keys():
            return self.extend_prop_map[p]
        return None


class EntityData:
    def __init__(self):
        self.prop: Prop = None
        self.biz_id: str = None
        self.name: str = ""
        self.description: str = ""
        self.type: str = None
        self.type_zh: str = None
        self.score = 1.0

    def get_properties_map_list_value(self):
        if self.prop is None:
            return {}
        return self.prop.get_properties_map_list_value()

    def to_show_id(self):
        if self.type in ['verify_op_result'] and self.description is not None and self.description != '':
            return f"{self.type_zh}[{self.name}]({self.description})"
        if self.name == self.biz_id:
            return f"{self.type_zh}[{self.name}]"
        return f"{self.type_zh}[{self.name}]({self.biz_id})"

    def to_json(self):
        return {
            "prop": {} if self.prop is None else self.prop.to_json(),
            "biz_id": self.biz_id,
            "name": self.name,
            "description": self.description,
            "type": self.type,
            "type_zh": self.type_zh
        }

    def get_attribute_value(self, p):
        if self.prop is None:
            return None
        return self.prop.get_prop_value(p)

    def merge_entity_data(self, other):
        if other.prop is not None:
            self.prop = other.prop
        if other.name is not None and other.name != '':
            self.name = other.name

        if other.description is not None and other.description != '':
            self.description = other.description

        if other.type is not None and other.type != '':
            self.type = other.type

        if other.type_zh is not None and other.type_zh != '':
            self.type_zh = other.type_zh

    def to_spo_list(self):
        spo_list = []
        spo_list.append(json.dumps({
            "s": self.name,
            "p": "归属类型",
            "o": self.type
        }, ensure_ascii=False))
        if self.prop is not None:
            for prop_key in self.prop.origin_prop_map.keys():
                if prop_key.startswith("_"):
                    continue
                if prop_key in ['id', 'name']:
                    continue
                spo_list.append(json.dumps({
                    "s": self.name,
                    "p": prop_key,
                    "o": self.prop.origin_prop_map[prop_key]
                }, ensure_ascii=False))
            for prop_key in self.prop.extend_prop_map.keys():
                spo_list.append(json.dumps({
                    "s": self.name,
                    "p": prop_key,
                    "o": self.prop.extend_prop_map[prop_key]
                }, ensure_ascii=False))
        return spo_list

    # def __repr__(self):
    #     return f"({self.name} [{self.type}] ({self.biz_id}) {self.description}"
    def __repr__(self):
        return f"{self.name} [{self.type}]"


def get_label_without_prefix(schema: SchemaUtils, label):
    if schema is None:
        return label
    return schema.get_label_without_prefix(label)


class RelationData:
    def __init__(self):
        self.prop: Prop = None
        self.from_id: str = None
        self.end_id: str = None
        self.from_entity: EntityData = None
        self.from_type: str = None
        self.from_alias = "s"
        self.end_type: str = None
        self.end_entity: EntityData = None
        self.end_alias = "o"
        self.type: str = None

    def get_spo_type(self):
        return f"{self.from_type}_{self.type}_{self.end_type}"

    def get_spo_show_id(self):
        return self.from_entity.to_show_id(), self.type, self.end_entity.to_show_id()

    def to_show_id(self):
        return f"{self.from_entity.to_show_id()} {self.type} {self.end_entity.to_show_id()}"

    def get_properties_map_list_value(self):
        if self.prop is None:
            return {}
        return self.prop.get_properties_map_list_value()

    def to_json(self):
        return {
            "prop": {} if self.prop is None else self.prop.to_json(),
            "from_id": self.from_id,
            "end_id": self.end_id,
            "from_entity_name": self.from_entity.name,
            "from_type": self.from_type,
            "end_entity_name": self.end_entity.name,
            "end_type": self.end_type,
            "type": self.type
        }

    def _get_entity_description(self, entity: EntityData):
        if entity is None:
            return None
        if entity.description is None or entity.description == '':
            return None
        if entity.type == "attribute":
            return None
        return entity.description

    def _get_entity_id(self, name: str, id: str):
        if id == name:
            return ""
        else:
            return f"[{id}]"

    def to_spo_list(self):
        spo_list = []
        rel = {
            "s": self.from_entity.name,
            "p": self.type,
            "o": self.end_entity.name
        }
        spo_list.append(json.dumps(rel, ensure_ascii=False))
        # prop
        if self.prop is not None:
            for prop_key in self.prop.origin_prop_map.keys():
                spo_list.append(json.dumps({
                    "s": rel,
                    "p": prop_key,
                    "o": self.prop.origin_prop_map[prop_key]
                }, ensure_ascii=False))
            for prop_key in self.prop.extend_prop_map.keys():
                spo_list.append(json.dumps({
                    "s": rel,
                    "p": prop_key,
                    "o": self.prop.extend_prop_map[prop_key]
                }, ensure_ascii=False))
        return spo_list

    def __repr__(self):
        from_entity_desc = self._get_entity_description(self.from_entity)
        from_entity_desc_str = "" if from_entity_desc is None else f"({from_entity_desc})"
        to_entity_desc = self._get_entity_description(self.end_entity)
        to_entity_desc_str = "" if to_entity_desc is None else f"({to_entity_desc})"
        return f"({self.from_entity.name}{from_entity_desc_str} {self.type} {self.end_entity.name}{to_entity_desc_str})"

    @staticmethod
    def from_dict(json_dict: dict, schema: SchemaUtils):
        rel = RelationData()

        rel.from_id = json_dict["__from_id__"]
        rel.from_type = get_label_without_prefix(schema, json_dict["__from_id_type__"])
        rel.end_id = json_dict["__to_id__"]
        rel.end_type = get_label_without_prefix(schema, json_dict["__to_id_type__"])
        rel.type = json_dict["__label__"]
        spo_label_name = f"{rel.from_type}_{rel.type}_{rel.end_type}"
        rel.prop = Prop.from_dict(json_dict, spo_label_name, schema)
        if schema is not None:
            if spo_label_name in schema.spo_en_zh.keys():
                rel.type = schema.get_spo_with_p(schema.spo_en_zh[spo_label_name])
        return rel

    def revert_spo(self):
        rel = RelationData()
        rel.from_id = self.end_id
        rel.from_type = self.end_type
        rel.from_entity = self.end_entity

        rel.end_id = self.from_id
        rel.end_type = self.from_type
        rel.end_entity = self.from_entity

        rel.type = self.type
        rel.prop = self.prop
        return rel

    @staticmethod
    def from_prop_value(s: EntityData, p: str, o: EntityData):
        rel = RelationData()
        rel.type = p

        rel.from_id = s.biz_id
        rel.from_type = s.type
        rel.from_entity = s

        rel.end_id = o.biz_id
        rel.end_type = o.type
        rel.end_entity = o
        return rel


class OneHopGraphData:
    def __init__(self, schema, alias_name):
        self.s_alias_name = alias_name
        self.s: EntityData = None
        self.in_relations: dict = {}
        self.out_relations: dict = {}
        self.schema = schema

    def to_graph_detail(self):
        s_po_map = {}
        prop_map = self.s.get_properties_map_list_value()
        # get out edge map
        for k in self.out_relations.keys():
            for rel in self.out_relations[k]:
                s, p, o = rel.get_spo_show_id()
                rel_prop_map = rel.get_properties_map_list_value()
                if len(rel_prop_map) > 0:
                    s_po_map[f"{s} {p} {o}"] = rel_prop_map
                if p in prop_map.keys():
                    prop_map[p].append(o)
                else:
                    prop_map[p] = [o]
                end_prop_map = rel.end_entity.get_properties_map_list_value()
                if o not in s_po_map.keys():
                    s_po_map[o] = end_prop_map

        s_po_map[self.s.to_show_id()] = prop_map

        for k in self.in_relations.keys():
            for rel in self.in_relations[k]:
                s, p, o = rel.get_spo_show_id()
                rel_prop_map = rel.get_properties_map_list_value()
                if len(rel_prop_map) > 0:
                    s_po_map[f"{s} {p} {o}"] = rel_prop_map
                start_prop_map = rel.from_entity.get_properties_map_list_value()
                if s not in s_po_map.keys():
                    s_po_map[s] = {
                        p: [o]
                    }
                else:
                    s_po_map[s].update({
                        p: [o]
                    })
                s_po_map[s].update(start_prop_map)
        return s_po_map

    def _schema_attr_en_to_zh(self, k):
        if self.schema is None:
            return k
        return self.schema.attr_en_zh.get(k)

    def get_s_all_attribute_spo(self):
        # spo_list = []
        attr_name_set_map = {}
        if self.s.prop is None:
            return attr_name_set_map
        if len(self.s.prop.origin_prop_map) > 0:
            for k in self.s.prop.origin_prop_map.keys():
                if k in ['id', 'name']:
                    continue
                if k.startswith("_"):
                    continue
                spo_list = []
                if isinstance(self.s.prop.origin_prop_map[k], list):
                    for v in self.s.prop.origin_prop_map[k]:
                        spo_list.append(str(v))
                else:
                    spo_list.append(str(self.s.prop.origin_prop_map[k]))
                attr_name_set_map[k] = spo_list
        if len(self.s.prop.extend_prop_map) > 0:
            for k in self.s.prop.extend_prop_map.keys():
                if k in ['id', 'name']:
                    continue
                if k.startswith("_"):
                    continue
                spo_list = []
                if isinstance(self.s.prop.origin_prop_map[k], list):
                    for v in self.s.prop.origin_prop_map[k]:
                        spo_list.append(str(v))
                else:
                    spo_list.append(str(self.s.prop.origin_prop_map[k]))
                attr_name_set_map[k] = spo_list
        return attr_name_set_map

    def get_s_all_attribute_name(self):
        attribute_name_set = []
        if self.s.prop is None:
            return attribute_name_set
        if len(self.s.prop.origin_prop_map) > 0:
            for k in self.s.prop.origin_prop_map.keys():
                attribute_name_set.append(self._schema_attr_en_to_zh(k))
        if len(self.s.prop.extend_prop_map) > 0:
            for k in self.s.prop.extend_prop_map.keys():
                attribute_name_set.append(k)
        return attribute_name_set

    def get_std_attribute_value(self, p):
        return self.s.get_attribute_value(p)

    def get_std_relation_value(self, p):
        relation_value_set = []
        if p in self.in_relations.keys():
            for rel in self.in_relations[p]:
                if "s" == self.s_alias_name:
                    relation_value_set.append(rel.revert_spo())
                else:
                    relation_value_set.append(rel)
        if p in self.out_relations.keys():
            for rel in self.out_relations[p]:
                if "o" == self.s_alias_name:
                    relation_value_set.append(rel.revert_spo())
                else:
                    relation_value_set.append(rel)
        return relation_value_set

    def get_all_relation_value(self):
        relation_value_set = []
        for p in self.in_relations.keys():
            for rel in self.in_relations[p]:
                if "s" == self.s_alias_name:
                    relation_value_set.append(rel.revert_spo())
                else:
                    relation_value_set.append(rel)
        for p in self.out_relations.keys():
            for rel in self.out_relations[p]:
                if "o" == self.s_alias_name:
                    relation_value_set.append(rel.revert_spo())
                else:
                    relation_value_set.append(rel)
        return relation_value_set

    def _prase_attribute_relation(self, std_p: str, attr_value: str):
        # new a RelationData
        prop_entity = EntityData()
        prop_entity.biz_id = attr_value
        prop_entity.name = attr_value
        prop_entity.type = "attribute"
        prop_entity.type_zh = "文本"

        return self._prase_entity_relation(std_p, prop_entity)

    def _prase_entity_relation(self, std_p: str, o_value: EntityData):
        s_entity = self.s
        o_entity = o_value
        if self.s_alias_name == "o":
            o_entity = self.s
            s_entity = o_value
        if o_value.description is None or o_value.description == '':
            o_value.description = f"{s_entity.name} {std_p} {o_entity.name}"
        return RelationData.from_prop_value(s_entity, std_p, o_entity)


    def get_std_attr_value_by_spo_text(self, p, spo_text):

        spo_list = []
        if self.s.prop is None:
            return spo_list
        if len(self.s.prop.origin_prop_map) > 0:
            for k in self.s.prop.origin_prop_map.keys():
                if k in ['id', 'name']:
                    continue
                if k.startswith("_"):
                    continue
                if isinstance(self.s.prop.origin_prop_map[k], list):
                    for v in self.s.prop.origin_prop_map[k]:
                        if spo_text == str(v):
                            spo_list.append(self._prase_attribute_relation(p, v))
                else:
                    if spo_text == str(self.s.prop.origin_prop_map[k]):
                        spo_list.append(self._prase_attribute_relation(p, spo_text))
        return spo_list

    def get_std_p_value_by_spo_text(self, p, spo_text):
        relation_value_set = []
        if p in self.in_relations.keys():
            for rel in self.in_relations[p]:
                if spo_text == str(rel).strip('(').strip(')'):
                    if "s" == self.s_alias_name:
                        relation_value_set.append(rel.revert_spo())
                    else:
                        relation_value_set.append(rel)
        if p in self.out_relations.keys():
            for rel in self.out_relations[p]:
                if spo_text == str(rel).strip('(').strip(')'):
                    if "o" == self.s_alias_name:
                        relation_value_set.append(rel.revert_spo())
                    else:
                        relation_value_set.append(rel)
        prop = self.s.prop.get_properties_map_list_value()
        if p in prop.keys():
            v_set = prop[p]
            for rel in v_set:
                relation_value_set.append(self._prase_attribute_relation(p, str(rel)))
        return relation_value_set


    def get_edge_en_to_zh(self, k):
        if self.schema is None:
            return k
        return self.schema.edge_en_zh.get(k, k)

    def get_s_all_relation_spo(self):
        # spo_list = []
        relation_name_set_map = {}
        if len(self.in_relations) > 0:
            for k in self.in_relations.keys():
                if k in ['similarity']:
                    continue
                spo_list = []
                for v in self.in_relations[k]:
                    spo_list.append(str(v).strip('(').strip(')'))
                relation_name_set_map[k] = spo_list
        if len(self.out_relations) > 0:
            for k in self.out_relations.keys():
                if k in ['similarity']:
                    continue
                spo_list = []
                for v in self.out_relations[k]:
                    spo_list.append(str(v).strip('(').strip(')'))
                relation_name_set_map[k] = spo_list
        return relation_name_set_map

    def get_s_all_relation_name(self):
        relation_name_set = []
        if len(self.in_relations) > 0:
            for k in self.in_relations.keys():
                relation_name_set.append(self.get_edge_en_to_zh(k))
        if len(self.out_relations) > 0:
            for k in self.out_relations.keys():
                relation_name_set.append(self.get_edge_en_to_zh(k))
        return relation_name_set


def copy_one_hop_graph_data(other: OneHopGraphData, alias_name: str):
    ret = OneHopGraphData(other.schema, alias_name)
    ret.s = other.s
    ret.in_relations = other.in_relations
    ret.out_relations = other.out_relations
    return ret


class KgGraph:
    def __init__(self):
        self.logic_form_base = {}
        self.start_node_alias_name = []
        self.start_node_name = []
        self.query_graph = {}
        self.nodes_alias = []
        self.edge_alias = []
        self.entity_map = {}
        self.edge_map = {}

    def merge_kg_graph(self, other, wo_intersect=True):
        self.nodes_alias = list(set(self.nodes_alias + other.nodes_alias))
        self.edge_alias = list(set(self.edge_alias + other.edge_alias))
        for n_alias in other.entity_map.keys():
            if n_alias in self.entity_map.keys():
                alias_entity_id_map = {}
                for n_data in other.entity_map[n_alias]:
                    alias_entity_id_map[n_data.biz_id] = n_data
                n_alias_merged_list = []
                for n_data in self.entity_map[n_alias]:
                    if n_data.biz_id in alias_entity_id_map.keys():
                        n_data.merge_entity_data(alias_entity_id_map[n_data.biz_id])
                        alias_entity_id_map.pop(n_data.biz_id)
                    n_alias_merged_list.append(n_data)
                n_alias_merged_list.extend(list(alias_entity_id_map.values()))
                self.entity_map[n_alias] = n_alias_merged_list
            else:
                self.entity_map[n_alias] = other.entity_map[n_alias]

        for e_alias in other.edge_map.keys():
            if e_alias in self.edge_map.keys():
                self.edge_map[e_alias] = self.edge_map[e_alias] + other.edge_map[e_alias]
            else:
                self.edge_map[e_alias] = other.edge_map[e_alias]
        for p in other.query_graph.keys():
            self.query_graph[p] = other.query_graph[p]
        if not wo_intersect:
            for p in other.query_graph.keys():
                self.query_graph[p] = other.query_graph[p]
                o_alias = self.query_graph[p]["o"]
                spo = self.get_all_relation_spo(o_alias)
                for s1, p1, o1 in spo:
                    if p1 == p:
                        continue
                    p_rels = self.edge_map[p] if p in self.edge_map.keys() else []
                    if p1 not in self.edge_map.keys():
                        continue
                    rels = self.edge_map[p1]
                    intersect_rels = []
                    for rel in rels:
                        for p_rel in p_rels:
                            if o1 == o_alias and rel.end_id == p_rel.end_id:
                                intersect_rels.append(rel)
                                break
                            if s1 == o_alias and rel.from_id == p_rel.end_id:
                                intersect_rels.append(rel)
                                break
                    self.edge_map[p1] = intersect_rels

    def to_edge_evidence(self):
        edge_map_str = []
        for k in self.edge_map.keys():
            rels = []
            for d in self.edge_map[k]:
                rels.append(str(d))
            edge_map_str.append(f"spo:{k} value:{','.join(rels)}")
        return edge_map_str

    def _entity_map_to_json(self):
        result_dict = {}
        for k in self.entity_map.keys():
            k_nodes = []
            for d in self.entity_map[k]:
                k_nodes.append(d.to_json())
            result_dict[k] = k_nodes
        return result_dict

    def _edge_map_to_json(self):
        result_dict = {}
        for k in self.edge_map.keys():
            rels = []
            for d in self.edge_map[k]:
                rels.append(d.to_json())
            result_dict[k] = rels
        return result_dict

    def to_answer_path(self):
        answer_path = []
        sp_o_map = {}
        for k in self.edge_map.keys():
            for d in self.edge_map[k]:
                s, p, o = d.get_spo_show_id()
                if (s, p) in sp_o_map.keys():
                    if o not in sp_o_map[(s, p)]:
                        sp_o_map[(s, p)].append(o)
                else:
                    sp_o_map[(s, p)] = [o]
        used_entities = []
        for k in sp_o_map.keys():
            answer_path.append({
                "s": k[0],
                "p": k[1],
                "o": sp_o_map[k]
            })
            used_entities.append(k[0])
            used_entities = used_entities + sp_o_map[k]
            used_entities = list(set(used_entities))
        return answer_path

    def get_all_entity_id(self):
        all_entity_id = []
        for k in self.edge_map.keys():
            for d in self.edge_map[k]:
                all_entity_id.append(d.from_id)
                all_entity_id.append(d.end_id)
        for k in self.entity_map.keys():
            for d in self.entity_map[k]:
                all_entity_id.append(d.biz_id)
        return list(set(all_entity_id))

    def get_all_entity(self):
        all_entity = []
        for k in self.edge_map.keys():
            for d in self.edge_map[k]:
                all_entity.append(d.from_entity)
                all_entity.append(d.end_entity)
        for k in self.entity_map.keys():
            for d in self.entity_map[k]:
                all_entity.append(d)
        return list(set(all_entity))

    def _graph_to_json(self):
        total_entity_map = {}
        edge_dict = {}
        has_entity = False
        for k in self.edge_map.keys():
            rels = []
            s_alias = self.query_graph[k]["s"]
            total_entity_map[s_alias] = []
            o_alias = self.query_graph[k]["o"]
            total_entity_map[o_alias] = []
            for d in self.edge_map[k]:
                has_entity = True
                rels.append(d.to_json())
                if d.from_alias == "s" and d.from_entity not in total_entity_map[s_alias]:
                    total_entity_map[s_alias].append(d.from_entity)
                if d.from_alias == "o" and d.from_entity not in total_entity_map[o_alias]:
                    total_entity_map[o_alias].append(d.from_entity)

                if d.end_alias == "s" and d.end_entity not in total_entity_map[s_alias]:
                    total_entity_map[s_alias].append(d.end_entity)
                if d.end_alias == "o" and d.end_entity not in total_entity_map[o_alias]:
                    total_entity_map[o_alias].append(d.end_entity)

            edge_dict[k] = rels
        if not has_entity:
            node_dict = self._entity_map_to_json()
        else:
            node_dict = {}
            for k in total_entity_map.keys():
                k_nodes = []
                for d in total_entity_map[k]:
                    k_nodes.append(d.to_json())
                node_dict[k] = k_nodes
        return edge_dict, node_dict

    def add_start_node_name_and_alias(self, alias_name, node_name_list):
        if alias_name not in self.start_node_alias_name:
            self.start_node_alias_name.append(alias_name)
        for n in node_name_list:
            if n not in self.start_node_name:
                self.start_node_name.append(n)

    def to_json(self):
        edge_dict, node_dict = self._graph_to_json()
        return {
            "query_graph": self.query_graph,
            "nodes_alias": self.nodes_alias,
            "edge_alias": self.edge_alias,
            "start_node_alias_name": list(set(self.start_node_alias_name)),
            "start_node_name": list(set(self.start_node_name)),
            "entity_map": node_dict,
            "edge_map": edge_dict
        }

    def to_edge_str(self):
        return "\n".join(self.to_edge_evidence())

    def to_node_evidence(self):
        node_map_str = []
        for k in self.entity_map.keys():
            k_node = []
            for d in self.entity_map[k]:
                k_node.append(str(d))
            node_map_str.append(f"alias {k}, value:{','.join(k_node)}")
        return node_map_str

    def to_node_str(self):
        return "\n".join(self.to_node_evidence())

    def to_evidence(self):
        if len(self.query_graph) == 0:
            return self.to_node_evidence()
        return self.to_edge_evidence()

    def to_spo(self):
        recored_ids = []
        spo_list = []
        for k in self.entity_map.keys():
            for d in self.entity_map[k]:
                spo_list = spo_list + d.to_spo_list()
                recored_ids.append(f"{d.type}_{d.biz_id}")

        for k in self.edge_map.keys():
            for d in self.edge_map[k]:
                spo_list = spo_list + d.to_spo_list()
                from_id_with_type = f"{d.from_type}_{d.from_id}"
                if from_id_with_type not in recored_ids and d.from_type != "attribute":
                    spo_list = spo_list + d.from_entity.to_spo_list()
                    recored_ids.append(from_id_with_type)

                end_id_with_type = f"{d.end_type}_{d.end_id}"
                if end_id_with_type not in recored_ids and d.end_type != "attribute":
                    spo_list = spo_list + d.end_entity.to_spo_list()
                    recored_ids.append(end_id_with_type)
        return spo_list

    def rmv_edge_ins(self, alias_name, alias_ins_set):
        if alias_name not in self.edge_map.keys():
            return
        update_rel_list = []
        for d in self.edge_map[alias_name]:
            if d not in alias_ins_set:
                update_rel_list.append(d)
        self.edge_map[alias_name] = update_rel_list

    def append_into_map(self, map: dict, key, value):
        old_list = []
        if key in map.keys():
            old_list = map[key]
        old_list.append(value)
        map[key] = old_list

    def rmv_node_ins(self, alias_name, alias_ins_set):
        if alias_name in self.entity_map.keys():
            entity_list = self.entity_map[alias_name]
            updated_entity_list = []
            for d in entity_list:
                if d.biz_id not in alias_ins_set:
                    updated_entity_list.append(d)
            self.entity_map[alias_name] = updated_entity_list

        # need generate with edge
        spo_set = self.get_all_relation_spo(alias_name)

        allowed_entity_dict = {}
        visited_p_dict = {}
        for s, p, o in spo_set:
            if p not in self.edge_map.keys():
                continue
            rel_list = []
            allowed_entity_dict[s.alias_name] = []
            allowed_entity_dict[o.alias_name] = []
            for rel in self.edge_map[p]:
                if (s.alias_name == alias_name and rel.from_id not in alias_ins_set) \
                        or (o.alias_name == alias_name and rel.end_id not in alias_ins_set):
                    rel_list.append(rel)
                    self.append_into_map(allowed_entity_dict, s.alias_name, rel.from_id)
                    self.append_into_map(allowed_entity_dict, o.alias_name, rel.end_id)
            self.edge_map[p] = rel_list
            visited_p_dict[p] = rel_list

        # 去除其余edge_map[p]中在filtered_entity_dict中的边
        for p in self.edge_map.keys():
            if p in visited_p_dict.keys():
                continue
            rel_list = []
            spo = self.query_graph[p]
            allowed_s = None
            allowed_o = None
            if spo["s"] in allowed_entity_dict.keys():
                allowed_s = allowed_entity_dict[spo["s"]]
            if spo["o"] in allowed_entity_dict.keys():
                allowed_o = allowed_entity_dict[spo["o"]]

            for rel in self.edge_map[p]:
                if allowed_s is not None and rel.from_id not in allowed_s:
                    continue
                if allowed_o is not None and rel.end_id not in allowed_o:
                    continue
                rel_list.append(rel)
            self.edge_map[p] = rel_list

    def rmv_ins(self, alias_name, alias_ins_set):
        if len(alias_ins_set) == 0:
            return
        if alias_name in self.nodes_alias:
            self.rmv_node_ins(alias_name.alias_name, alias_ins_set)
        elif alias_name in self.edge_alias:
            self.rmv_edge_ins(alias_name.alias_name, alias_ins_set)

    def rmv_node_alias(self, alias):
        if alias not in self.nodes_alias:
            return
        self.entity_map.pop(alias)
        self.nodes_alias.remove(alias)

    def get_all_relation_spo(self, alias):
        res = []
        for p in self.query_graph.keys():
            spo = self.query_graph[p]
            if spo["o"] == alias or spo["s"] == alias:
                res.append((spo["s"], spo["p"], spo["o"]))
        return res

    def get_entity_by_alias(self, alias):
        if alias in self.nodes_alias:
            if alias in self.entity_map.keys():
                return self.entity_map[alias]
            else:
                # need generate with edge
                spo_set = self.get_all_relation_spo(alias)
                ret_entity = []
                for s, p, o in spo_set:
                    if p not in self.edge_map.keys():
                        continue
                    for rel in self.edge_map[p]:
                        if s == alias:
                            ret_entity.append(rel.from_entity)
                        if o == alias:
                            ret_entity.append(rel.end_entity)
                if len(ret_entity) == 0:
                    return None
                return ret_entity
        if alias in self.edge_map.keys():
            return self.edge_map[alias]
        return None



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os

import bs4.element
import markdown
from bs4 import BeautifulSoup, Tag
from typing import List, Type
import logging
import re
import requests
import pandas as pd
from io import StringIO
from tenacity import stop_after_attempt, retry

from kag.interface.builder import SourceReaderABC
from kag.builder.model.chunk import Chunk, ChunkTypeEnum
from knext.common.base.runnable import Output, Input
from kag.builder.prompt.analyze_table_prompt import AnalyzeTablePrompt


class MarkDownReader(SourceReaderABC):
    """
    A class for reading MarkDown files, inheriting from `SourceReader`.
    Supports converting MarkDown data into a list of Chunk objects.

    Args:
        cut_depth (int): The depth of cutting, determining the level of detail in parsing. Default is 1.
    """

    ALL_LEVELS = [f"h{x}" for x in range(1, 7)]
    TABLE_CHUCK_FLAG = "<<<table_chuck>>>"

    def __init__(self, cut_depth: int = 1, **kwargs):
        super().__init__(**kwargs)
        self.cut_depth = int(cut_depth)
        self.llm_module = kwargs.get("llm_module", None)
        self.analyze_table_prompt = AnalyzeTablePrompt(language="zh")
        self.analyze_img_prompt = AnalyzeTablePrompt(language="zh")

    @property
    def input_types(self) -> Type[Input]:
        return str

    @property
    def output_types(self) -> Type[Output]:
        return Chunk

    def to_text(self, level_tags):
        """
        Converts parsed hierarchical tags into text content.

        Args:
            level_tags (list): Parsed tags organized by Markdown heading levels and other tags.

        Returns:
            str: Text content derived from the parsed tags.
        """
        content = []
        for item in level_tags:
            if isinstance(item, list):
                content.append(self.to_text(item))
            else:
                header, tag = item
                if not isinstance(tag, Tag):
                    continue
                elif tag.name in self.ALL_LEVELS:
                    content.append(
                        f"{header}-{tag.text}" if len(header) > 0 else tag.text
                    )
                else:
                    content.append(self.tag_to_text(tag))
        return "\n".join(content)

    def tag_to_text(self, tag: bs4.element.Tag):
        """
        将html tag转换为text
        如果是table，输出markdown，添加表格标记，方便后续构建Chunk
        :param tag:
        :return:
        """
        if tag.name == "table":
            try:
                html_table = str(tag)
                table_df = pd.read_html(html_table)[0]
                return f"{self.TABLE_CHUCK_FLAG}{table_df.to_markdown(index=False)}{self.TABLE_CHUCK_FLAG}"
            except:
                logging.warning("parse table tag to text error", exc_info=True)
        return tag.text

    @retry(stop=stop_after_attempt(5))
    def analyze_table(self, table,analyze_mathod="human"):
        if analyze_mathod == "llm":
            if self.llm_module == None:
                logging.INFO("llm_module is None, cannot use analyze_table")
                return table
            variables = {
                "table": table
            }
            response = self.llm_module.invoke(
                variables = variables,
                prompt_op = self.analyze_table_prompt,
                with_json_parse=False
            )
            if response is  None or response == "" or response == []:
                raise Exception("llm_module return None")
            return response
        else:
            from io import StringIO
            import pandas as pd
            try:
                df = pd.read_html(StringIO(table))[0]
            except Exception as e:
                logging.warning(f"analyze_table error: {e}")
                return table
            content = ""
            for index, row in df.iterrows():
                content+=f"第{index+1}行的数据如下:"
                for col_name, value in row.items():
                    content+=f"{col_name}的值为{value}，"
                content+='\n'
            return content

    
    @retry(stop=stop_after_attempt(5))
    def analyze_img(self, img_url):
        response = requests.get(img_url)
        response.raise_for_status()
        image_data = response.content

        pass

    def replace_table(self, content: str):
        pattern = r"<table[^>]*>([\s\S]*?)<\/table>"
        for match in re.finditer(pattern, content):
            table = match.group(0)
            table = self.analyze_table(table)
            content = content.replace(match.group(1), table)
        return content

    def replace_img(self, content: str):
        pattern = r"<img[^>]*src=[\"\']([^\"\']*)[\"\']"
        for match in re.finditer(pattern, content):
            img_url = match.group(1)
            img_msg = self.analyze_img(img_url)
            content = content.replace(match.group(0), img_msg)
        return content

    def extract_table(self, level_tags, header=""):
        """
        Extracts tables from the parsed hierarchical tags along with their headers.

        Args:
            level_tags (list): Parsed tags organized by Markdown heading levels and other tags.
            header (str): Current header text being processed.

        Returns:
            list: A list of tuples, each containing the table's header, context text, and the table tag.
        """
        tables = []
        for idx, item in enumerate(level_tags):
            if isinstance(item, list):
                tables += self.extract_table(item, header)
            else:
                tag = item[1]
                if not isinstance(tag, Tag):
                    continue
                if tag.name in self.ALL_LEVELS:
                    header = f"{header}-{tag.text}" if len(header) > 0 else tag.text

                if tag.name == "table":
                    if idx - 1 >= 0:
                        context = level_tags[idx - 1]
                        if isinstance(context, tuple):
                            tables.append((header, context[1].text, tag))
                    else:
                        tables.append((header, "", tag))
        return tables

    def parse_level_tags(
            self,
            level_tags: list,
            level: str,
            parent_header: str = "",
            cur_header: str = "",
    ):
        """
        Recursively parses level tags to organize them into a structured format.

        Args:
            level_tags (list): A list of tags to be parsed.
            level (str): The current level being processed.
            parent_header (str): The header of the parent tag.
            cur_header (str): The header of the current tag.

        Returns:
            list: A structured representation of the parsed tags.
        """
        if len(level_tags) == 0:
            return []
        output = []
        prefix_tags = []
        while len(level_tags) > 0:
            tag = level_tags[0]
            if tag.name in self.ALL_LEVELS:
                break
            else:
                prefix_tags.append((parent_header, level_tags.pop(0)))
        if len(prefix_tags) > 0:
            output.append(prefix_tags)

        cur = []
        while len(level_tags) > 0:
            tag = level_tags[0]
            if tag.name not in self.ALL_LEVELS:
                cur.append((parent_header, level_tags.pop(0)))
            else:

                if tag.name > level:
                    cur += self.parse_level_tags(
                        level_tags,
                        tag.name,
                        f"{parent_header}-{cur_header}"
                        if len(parent_header) > 0
                        else cur_header,
                        tag.name,
                    )
                elif tag.name == level:
                    if len(cur) > 0:
                        output.append(cur)
                    cur = [(parent_header, level_tags.pop(0))]
                    cur_header = tag.text
                else:
                    if len(cur) > 0:
                        output.append(cur)
                    return output
        if len(cur) > 0:
            output.append(cur)
        return output

    def cut(self, level_tags, cur_level, final_level):
        """
        Cuts the provided level tags into chunks based on the specified levels.

        Args:
            level_tags (list): A list of tags to be cut.
            cur_level (int): The current level in the hierarchy.
            final_level (int): The final level to which the tags should be cut.

        Returns:
            list: A list of cut chunks.
        """
        output = []
        if cur_level == final_level:
            cur_prefix = []
            for sublevel_tags in level_tags:
                if (
                        isinstance(sublevel_tags, tuple)
                ):
                    cur_prefix.append(self.to_text([sublevel_tags,]))
                else:
                    break
            cur_prefix = "\n".join(cur_prefix)

            if len(cur_prefix) > 0:
                output.append(cur_prefix)
            for sublevel_tags in level_tags:
                if isinstance(sublevel_tags, list):
                    output.append(cur_prefix + "\n" + self.to_text(sublevel_tags))
            return output
        else:
            cur_prefix = []
            for sublevel_tags in level_tags:
                if (
                        isinstance(sublevel_tags, tuple)
                ):
                    cur_prefix.append(sublevel_tags[1].text)
                else:
                    break
            cur_prefix = "\n".join(cur_prefix)
            if len(cur_prefix) > 0:
                output.append(cur_prefix)

            for sublevel_tags in level_tags:
                if isinstance(sublevel_tags, list):
                    output += self.cut(sublevel_tags, cur_level + 1, final_level)
            return output

    def solve_content(self, id: str, title: str, content: str, **kwargs) -> List[Output]:
        """
        Converts Markdown content into structured chunks.

        Args:
            id (str): An identifier for the content.
            title (str): The title of the content.
            content (str): The Markdown formatted content to be processed.

        Returns:
            List[Output]: A list of processed content chunks.
        """
        html_content = markdown.markdown(
            content, extensions=["markdown.extensions.tables"]
        )
        # html_content = self.replace_table(html_content)
        soup = BeautifulSoup(html_content, "html.parser")
        if soup is None:
            raise ValueError("The MarkDown file appears to be empty or unreadable.")

        top_level = None
        for level in self.ALL_LEVELS:
            tmp = soup.find_all(level)
            if len(tmp) > 0:
                top_level = level
                break
        if top_level is None:
            chunk = Chunk(
                id=Chunk.generate_hash_id(str(id)),
                name=title,
                content=soup.text,
                ref=kwargs.get("ref", ""),
            )
            return [chunk]
        tags = [tag for tag in soup.children if isinstance(tag, Tag)]

        level_tags = self.parse_level_tags(tags, top_level)
        cutted = self.cut(level_tags, 0, self.cut_depth)

        chunks = []

        for idx, content in enumerate(cutted):
            chunk = None
            if self.TABLE_CHUCK_FLAG in content:
                chunk = self.get_table_chuck(content, title, id, idx)
                chunk.ref = kwargs.get("ref", "")
            else:
                chunk = Chunk(
                    id=Chunk.generate_hash_id(f"{id}#{idx}"),
                    name=f"{title}#{idx}",
                    content=content,
                    ref=kwargs.get("ref", ""),
                )
            chunks.append(chunk)
        return chunks

    def get_table_chuck(self, table_chunk_str: str, title: str, id: str, idx: int) -> Chunk:
        """
        convert table chunk
        :param table_chunk_str:
        :return:
        """
        table_chunk_str = table_chunk_str.replace("\\N", "")
        pattern = f"{self.TABLE_CHUCK_FLAG}(.*){self.TABLE_CHUCK_FLAG}"
        matches = re.findall(pattern, table_chunk_str, re.DOTALL)
        if not matches or len(matches) <= 0:
            # 找不到表格信息，按照Text Chunk处理
            return Chunk(
                id=Chunk.generate_hash_id(f"{id}#{idx}"),
                name=f"{title}#{idx}",
                content=table_chunk_str,
            )
        table_markdown_str = matches[0]
        html_table_str = markdown.markdown(table_markdown_str, extensions=["markdown.extensions.tables"])
        try:
            df = pd.read_html(html_table_str)[0]
        except Exception as e:
            logging.warning(f"get_table_chuck error: {e}")
            df = pd.DataFrame()

        # 确认是表格Chunk，去除内容中的TABLE_CHUCK_FLAG
        replaced_table_text = re.sub(pattern, f'\n{table_markdown_str}\n', table_chunk_str, flags=re.DOTALL)
        return Chunk(
            id=Chunk.generate_hash_id(f"{id}#{idx}"),
            name=f"{title}#{idx}",
            content=replaced_table_text,
            type=ChunkTypeEnum.Table,
            csv_data=df.to_csv(index=False),
        )

    def invoke(self, input: Input, **kwargs) -> List[Output]:
        """
        Processes a Markdown file and returns its content as structured chunks.

        Args:
            input (Input): The path to the Markdown file.
            **kwargs: Additional keyword arguments.

        Returns:
            List[Output]: A list of processed content chunks.
        """
        file_path: str = input

        if not file_path.endswith(".md"):
            raise ValueError(f"Please provide a markdown file, got {file_path}")

        if not os.path.isfile(file_path):
            raise FileNotFoundError(f"The file {file_path} does not exist.")

        with open(file_path, "r") as reader:
            content = reader.read()

        basename, _ = os.path.splitext(os.path.basename(file_path))

        chunks = self.solve_content(input, basename, content)
        return chunks



import os
from typing import List

import numpy as np

from kag.common.vectorizer import Vectorizer


def cosine_similarity(vector1, vector2):
    cosine = np.dot(vector1, vector2) / (np.linalg.norm(vector1) * np.linalg.norm(vector2))
    return cosine

def split_list(input_list, max_length=30):
    """
    Splits a list into multiple sublists where each sublist has a maximum length of max_length.

    :param input_list: The original list to be split
    :param max_length: The maximum length of each sublist
    :return: A list containing multiple sublists
    """
    return [input_list[i:i + max_length] for i in range(0, len(input_list), max_length)]


class TextSimilarity:
    def __init__(self, vec_config=None):
        if vec_config is None:
            vec_config = eval(os.getenv("KAG_VECTORIZER"))
            if vec_config is None:
                message = "vectorizer config is required"
                raise RuntimeError(message)
        self._vectorizer: Vectorizer = Vectorizer.from_config(vec_config)

        self.cached_embs = {}

    def sentence_encode(self, sentences, is_cached=False):
        if isinstance(sentences, str):
            return self._vectorizer.vectorize(sentences)
        if not isinstance(sentences, list):
            return []
        if len(sentences) == 0:
            return []
        ready_list = split_list(sentences)
        ret = []
        for text_list in ready_list:
            tmp_map = {}
            need_call_emb_text = []
            for text in text_list:
                if text in self.cached_embs:
                    tmp_map[text] = self.cached_embs[text]
                else:
                    need_call_emb_text.append(text)
            if len(need_call_emb_text) > 0:
                emb_res = self._vectorizer.vectorize(need_call_emb_text)
                for text, text_emb in zip(need_call_emb_text, emb_res):
                    tmp_map[text] = text_emb
                    if is_cached:
                        self.cached_embs[text] = text_emb
            for text in text_list:
                ret.append(tmp_map[text])
        return ret

    def text_sim_result(self, mention, candidates: List[str], topk=1, low_score=0.63):
        '''
        output: [(candi_name, candi_score),...]
        '''
        if mention is None:
            return []
        mention_emb = self.sentence_encode(mention)
        candidates = [cand for cand in candidates if cand is not None and cand.strip() != '']
        if len(candidates) == 0:
            return []
        candidates_emb = self.sentence_encode(candidates)
        candidates_dis = {}
        for candidate, candidate_emb in zip(candidates, candidates_emb):
            cosine = cosine_similarity(np.array(mention_emb), np.array(candidate_emb))
            if cosine < low_score:
                continue
            candidates_dis[candidate] = cosine
        candidates_dis = sorted(candidates_dis.items(), key=lambda x:x[-1], reverse=True)
        candis = candidates_dis[:topk]
        return candis

    def text_type_sim(self, mention, candidates, topk=1):
        '''
        output: [(candi_name, candi_score),...]
        '''
        res = self.text_sim_result(mention, candidates, topk)
        if len(res) == 0:
            return [('Entity', 1.)]
        return res



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import json
from typing import Optional, List

from kag.common.base.prompt_op import PromptOp


class OpenIEEntitystandardizationdPrompt(PromptOp):
    template_en = """
{
    "instruction": "The `input` field contains a user provided context. The `named_entities` field contains extracted named entities from the context, which may be unclear abbreviations, aliases, or slang. To eliminate ambiguity, please attempt to provide the official names of these entities based on the context and your own knowledge. Note that entities with the same meaning can only have ONE official name. Please respond in the format of a single JSONArray string without any explanation, as shown in the `output` field of the provided example.",
    "example": {
        "input": "American History\nWhen did the political party that favored harsh punishment of southern states after the Civil War, gain control of the House? Republicans regained control of the chamber they had lost in the 2006 midterm elections.",
        "named_entities": [
            {"entity": "American", "category": "GeographicLocation"},
            {"entity": "political party", "category": "Organization"},
            {"entity": "southern states", "category": "GeographicLocation"},
            {"entity": "Civil War", "category": "Keyword"},
            {"entity": "House", "category": "Organization"},
            {"entity": "Republicans", "category": "Organization"},
            {"entity": "chamber", "category": "Organization"},
            {"entity": "2006 midterm elections", "category": "Date"}
        ],
        "output": [
            {
                "entity": "American",
                "category": "GeographicLocation",
                "official_name": "United States of America"
            },
            {
                "entity": "political party",
                "category": "Organization",
                "official_name": "Radical Republicans"
            },
            {
                "entity": "southern states",
                "category": "GeographicLocation",
                "official_name": "Confederacy"
            },
            {
                "entity": "Civil War",
                "category": "Keyword",
                "official_name": "American Civil War"
            },
            {
                "entity": "House",
                "category": "Organization",
                "official_name": "United States House of Representatives"
            },
            {
                "entity": "Republicans",
                "category": "Organization",
                "official_name": "Republican Party"
            },
            {
                "entity": "chamber",
                "category": "Organization",
                "official_name": "United States House of Representatives"
            },
            {
                "entity": "midterm elections",
                "category": "Date",
                "official_name": "United States midterm elections"
            }
        ]
    },
    "input": "$input",
    "named_entities": $named_entities
}
    """

    template_zh = """
{
    "instruction": "input字段包含用户提供的上下文。命名实体字段包含从上下文中提取的命名实体，这些可能是含义不明的缩写、别名或俚语。为了消除歧义，请尝试根据上下文和您自己的知识提供这些实体的官方名称。请注意，具有相同含义的实体只能有一个官方名称。请按照提供的示例中的输出字段格式，以单个JSONArray字符串形式回复，无需任何解释。",
    "example": {
        "input": "烦躁不安、语妄、失眠酌用镇静药，禁用抑制呼吸的镇静药。\n3.并发症的处理经抗菌药物治疗后，高热常在24小时内消退，或数日内逐渐下降。\n若体温降而复升或3天后仍不降者，应考虑SP的肺外感染，如腋胸、心包炎或关节炎等。治疗：接胸腔压力调节管＋吸引机负压吸引水瓶装置闭式负压吸引宜连续，如经12小时后肺仍未复张，应查找原因。",
        "named_entities": [
            {"entity": "烦躁不安", "category": "Symptom"},
            {"entity": "语妄", "category": "Symptom"},
            {"entity": "失眠", "category": "Symptom"},
            {"entity": "镇静药", "category": "Medicine"},
            {"entity": "肺外感染", "category": "Disease"},
            {"entity": "胸腔压力调节管", "category": "MedicalEquipment"},
            {"entity": "吸引机负压吸引水瓶装置", "category": "MedicalEquipment"},
            {"entity": "闭式负压吸引", "category": "SurgicalOperation"}
        ],
        "output": [
            {"entity": "烦躁不安", "category": "Symptom", "official_name": "焦虑不安"},
            {"entity": "语妄", "category": "Symptom", "official_name": "谵妄"},
            {"entity": "失眠", "category": "Symptom", "official_name": "失眠症"},
            {"entity": "镇静药", "category": "Medicine", "official_name": "镇静剂"},
            {"entity": "肺外感染", "category": "Disease", "official_name": "肺外感染"},
            {"entity": "胸腔压力调节管", "category": "MedicalEquipment", "official_name": "胸腔引流管"},
            {"entity": "吸引机负压吸引水瓶装置", "category": "MedicalEquipment", "official_name": "负压吸引装置"},
            {"entity": "闭式负压吸引", "category": "SurgicalOperation", "official_name": "闭式负压引流"}
        ]
    },
    "input": $input,
    "named_entities": $named_entities,
}    
    """

    def __init__(self, language: Optional[str] = "en"):
        super().__init__(language)

    @property
    def template_variables(self) -> List[str]:
        return ["input", "named_entities"]

    def parse_response(self, response: str, **kwargs):

        rsp = response
        if isinstance(rsp, str):
            rsp = json.loads(rsp)
        if isinstance(rsp, dict) and "output" in rsp:
            rsp = rsp["output"]
        if isinstance(rsp, dict) and "named_entities" in rsp:
            standardized_entity = rsp["named_entities"]
        else:
            standardized_entity = rsp
        entities_with_offical_name = set()
        merged = []
        entities = kwargs.get("named_entities", [])
        for entity in standardized_entity:
            merged.append(entity)
            entities_with_offical_name.add(entity["entity"])
        # in case llm ignores some entities
        for entity in entities:
            if entity["entity"] not in entities_with_offical_name:
                entity["official_name"] = entity["entity"]
                merged.append(entity)
        return merged



import os
import time
from typing import List

import numpy as np

from kag.common.retriever import DefaultRetriever
from kag.solver.logic.core_modules.common.one_hop_graph import EntityData
from kag.solver.logic.core_modules.common.text_sim_by_vector import TextSimilarity, cosine_similarity
from kag.solver.logic.core_modules.retriver.retrieval_spo import logger
from knext.project.client import ProjectClient


class LFChunkRetriever(DefaultRetriever):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        vectorizer_config = eval(os.getenv("KAG_VECTORIZER", "{}"))
        if self.host_addr and self.project_id:
            config = ProjectClient(host_addr=self.host_addr, project_id=self.project_id).get_config(self.project_id)
            vectorizer_config.update(config.get("vectorizer", {}))
        self.text_sim = TextSimilarity(vec_config=vectorizer_config)

    def rerank(self, queries: List[str], passages: List[str]):
        if not isinstance(queries, list):
            queries = [queries]
        if len(passages) == 0:
            return []
        queries = list(set(queries))
        rank_scores = np.array([1 / (1 + i) for i in range(len(passages))])
        passage_scores = np.zeros(len(passages)) + rank_scores
        passages_embs = self.text_sim.sentence_encode(passages, is_cached=True)

        for query in queries:
            query_emb = self.text_sim.sentence_encode(query)
            scores = [cosine_similarity(np.array(query_emb), np.array(passage_emb)) for passage_emb in passages_embs]
            sorted_idx = np.argsort(-np.array(scores))
            for rank, passage_id in enumerate(sorted_idx):
                passage_scores[passage_id] += rank_scores[rank]

        merged_sorted_idx = np.argsort(-passage_scores)

        new_passages = [passages[x] for x in merged_sorted_idx]
        return new_passages[:10]

    def recall_docs(self, query: str, top_k=5, **kwargs):
        all_related_entities = kwargs.get('related_entities', None)
        query_ner_dict = kwargs.get('query_ner_dict', None)
        req_id = kwargs.get('req_id', '')
        if all_related_entities is None:
            return super().recall_docs(query, top_k, **kwargs)
        return self.recall_docs_by_entities(query, all_related_entities, top_k, req_id, query_ner_dict)

    def get_std_ner_by_query(self, query: str):
        """
        Retrieves standardized Named Entity Recognition (NER) results based on the input query.

        Parameters:
        query (str): The input query string.

        Returns:
        dict: A dictionary containing standardized entity names and types along with their scores.
        """
        entities = self.named_entity_recognition(query)
        entities_with_official_name = self.named_entity_standardization(
            query, entities
        )

        query_ner_list = entities_with_official_name
        try:
            names = []
            for x in query_ner_list:
                x_type = x.get("category", "Others").lower()
                if x_type in ["works", "person", "other"]:
                    names.append(x["entity"])
                else:
                    names.append(x["official_name"])
        except:
            names = [x["entity"] for x in query_ner_list if "entity" in x]
        types = [x["category"] for x in query_ner_list if "category" in x]

        query_ner_list = names
        query_ner_type_list = types

        top_phrases = []
        top_phrase_scores = []

        if len(query_ner_list) > 0:
            (top_phrases, top_phrase_scores) = self.match_entities(
                dict(zip(query_ner_list, query_ner_type_list))
            )
        return_data = {}
        for i in range(0, len(top_phrases)):
            phrases = top_phrases[i]
            phrases["score"] = top_phrase_scores[i]
            return_data[f"{phrases['name']}_{phrases['type']}"] = phrases
        return return_data

    def recall_docs_by_entities(self, query: str, all_related_entities: List[EntityData], top_k=10,
                                req_id='', query_ner_dict: dict = None):
        def convert_entity_data_to_ppr_cand(related_entities: List[EntityData]):
            ret_ppr_candis = {}
            for e in related_entities:
                k = f"{e.name}_{e.type}"
                ret_ppr_candis[k] = {
                    'name': e.name,
                    'type': e.type,
                    'score': e.score
                }
            return ret_ppr_candis

        start_time = time.time()
        ner_cands = self.get_std_ner_by_query(query)
        try:
            kg_cands = convert_entity_data_to_ppr_cand(all_related_entities)
        except Exception as e:
            kg_cands = {}
            logger.warning(f"{req_id} {query} generate logic form failed {str(e)}", exc_info=True)
        for k, v in ner_cands.items():
            if k in kg_cands.keys():
                if v['score'] > kg_cands[k]['score']:
                    kg_cands[k]['score'] = v['score']
            else:
                kg_cands[k] = v
        if query_ner_dict is not None:
            for k, v in query_ner_dict.items():
                if k in kg_cands.keys():
                    if v['score'] > kg_cands[k]['score']:
                        kg_cands[k]['score'] = v['score']
                else:
                    kg_cands[k] = v

        matched_entities = []
        matched_entities_scores = []
        for _, v in kg_cands.items():
            matched_entities.append(v)
            matched_entities_scores.append(v['score'])
        logger.info(f"{req_id} kgpath ner cost={time.time() - start_time}")

        start_time = time.time()
        if len(matched_entities) == 0:
            combined_scores = self.calculate_sim_scores(query, top_k * 20)
            logger.info(f"{req_id} only get_dpr_scores cost={time.time() - start_time}")
        elif (
                matched_entities_scores and np.min(matched_entities_scores) > self.pagerank_threshold
        ):  # high confidence in named entities
            combined_scores = self.calculate_pagerank_scores(matched_entities)
        else:
            # Run Personalized PageRank (PPR) or other Graph Algorithm Doc Scores
            pagerank_scores = self.calculate_pagerank_scores(matched_entities)
            logger.info(f"{req_id} only get_ppr_scores cost={time.time() - start_time}")
            start_time = time.time()
            sim_doc_scores = self.calculate_sim_scores(query, top_k * 20)
            logger.info(f"{req_id} only get_dpr_scores cost={time.time() - start_time}")

            combined_scores = self.calculate_combined_scores(sim_doc_scores, pagerank_scores)

        # Return ranked docs and ranked scores
        sorted_doc_ids = sorted(
                combined_scores.items(), key=lambda item: item[1], reverse=True
            )
        logger.debug(f"kgpath chunk recall cost={time.time() - start_time}")
        return self.get_all_docs_by_id(query, sorted_doc_ids, top_k)


from string import Template
from typing import List

from kag.common.base.prompt_op import PromptOp


class SolveQuestionWithOutSPO(PromptOp):

    template_zh = """请根据检索到的相关文档回答问题“$question”，并结合历史信息进行综合分析。
要求：
1.尽可能直接回答问题，不包括任何其他信息。
2.不要重复问题的内容。
3.根据提供的信息生成答案。如果可能有多个答案，请生成所有答案。
4.如果没有合适的答案，请回答“I don't know”。
历史：
$history
文档：
$docs

答案：
"""

    template_en = """Please answer the question `$question` based on the retrieved relevant documents, and combine historical information for comprehensive analysis.
Requirement:
1. Answer the question as directly as possible, without including any other information.
2. Do not repeat the content of the question.
3. Generate answers based on the provided information. If multiple answers are possible, generate all of them.
4. If there is no suitable answer, answer 'I don't know'.

history:
$history

docs:
$docs

answer:
"""

    def __init__(self, language: str):
        super().__init__(language)

    @property
    def template_variables(self) -> List[str]:
        return ["history", "question",  "docs"]

    def parse_response(self, response: str, **kwargs):
        return response



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os
import re
from typing import List, Sequence, Type, Union

from langchain_community.document_loaders import PyPDFLoader
import pdfminer.layout

from kag.builder.model.chunk import Chunk
from kag.interface.builder import SourceReaderABC
from knext.common.base.runnable import Input, Output
from kag.builder.prompt.outline_prompt import OutlinePrompt

from pdfminer.high_level import extract_text
from pdfminer.high_level import extract_pages
from pdfminer.layout import LTTextContainer, LTPage
from pdfminer.pdfparser import PDFParser
from pdfminer.pdfdocument import PDFDocument
from pdfminer.layout import LAParams,LTTextBox
from pdfminer.pdfpage import PDFPage
from pdfminer.pdfparser import PDFParser
from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.converter import PDFPageAggregator
from pdfminer.pdfpage import PDFTextExtractionNotAllowed
import pdfminer


import logging

logger = logging.getLogger(__name__)


class PDFReader(SourceReaderABC):
    """
    A PDF reader class that inherits from SourceReader.

    Attributes:
        if_split (bool): Whether to split the content by pages. Default is False.
        use_pypdf (bool): Whether to use PyPDF2 for processing PDF files. Default is True.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.split_level = kwargs.get("split_level", 3)
        self.split_using_outline = kwargs.get("split_using_outline", True)
        self.outline_flag = True
        self.llm = self._init_llm()
        language = os.getenv("KAG_PROMPT_LANGUAGE", "zh")
        self.prompt = OutlinePrompt(language)


    @property
    def input_types(self) -> Type[Input]:
        return str

    @property
    def output_types(self) -> Type[Output]:
        return Chunk
    
    def outline_chunk(self, chunk: Union[Chunk, List[Chunk]],basename) -> List[Chunk]:
        if isinstance(chunk, Chunk):
            chunk = [chunk]
        outlines = []
        for c in chunk:
            outline = self.llm.invoke({"input": c.content}, self.prompt)
            outlines.extend(outline)
        content = "\n".join([c.content for c in chunk])
        chunks = self.sep_by_outline(content, outlines,basename)
        return chunks
    
    def sep_by_outline(self,content,outlines,basename):
        position_check = []
        for outline in outlines:
            start = content.find(outline)
            position_check.append((outline,start))
        chunks = []
        for idx,pc in enumerate(position_check):
            chunk = Chunk(
                id = Chunk.generate_hash_id(f"{basename}#{pc[0]}"),
                name=f"{basename}#{pc[0]}",
                content=content[pc[1]:position_check[idx+1][1] if idx+1 < len(position_check) else len(position_check)],
            )
            chunks.append(chunk)
        return chunks

        

    @staticmethod
    def _process_single_page(
        page: str,
        watermark: str,
        remove_header: bool = False,
        remove_footnote: bool = False,
        remove_lists: List[str] = None,
    ) -> list:
        """
        Processes a single page of text, removing headers, footnotes, watermarks, and specified lists.

        Args:
            page (str): The text content of a single page.
            watermark (str): The watermark text to be removed.
            remove_header (bool): Whether to remove the header. Default is False.
            remove_footnote (bool): Whether to remove the footnote. Default is False.
            remove_lists (List[str]): A list of strings to be removed. Default is None.

        Returns:
            list: A list of processed text lines.
        """
        lines = page.split("\n")
        if remove_header and len(lines) > 0:
            lines = lines[1:]
        if remove_footnote and len(lines) > 0:
            lines = lines[:-1]

        cleaned = [line.strip().replace(watermark, "") for line in lines]

        if remove_lists is None:
            return cleaned
        for s in remove_lists:
            cleaned = [line.strip().replace(s, "") for line in cleaned]

        return cleaned

    @staticmethod
    def _extract_text_from_page(page_layout: LTPage) -> str:
        """
        Extracts text from a given page layout.

        Args:
            page_layout (LTPage): The layout of the page containing text elements.

        Returns:
            str: The extracted text.
        """
        text = ""
        for element in page_layout:
            if isinstance(element, LTTextContainer):
                text += element.get_text()
        return text

    def invoke(self, input: str, **kwargs) -> Sequence[Output]:
        """
        Processes a PDF file, splitting or extracting content based on configuration.

        Args:
            input (str): The path to the PDF file.
            **kwargs: Additional keyword arguments, such as `clean_list`.

        Returns:
            Sequence[Output]: A sequence of processed outputs.

        Raises:
            ValueError: If the file is not a PDF file or the content is empty/unreadable.
            FileNotFoundError: If the file does not exist.
        """
        if not input.endswith(".pdf"):
            raise ValueError(f"Please provide a pdf file, got {input}")

        if not os.path.isfile(input):
            raise FileNotFoundError(f"The file {input} does not exist.")


        self.fd = open(input, "rb")
        self.parser = PDFParser(self.fd)
        self.document = PDFDocument(self.parser)
        chunks = []
        basename, _ = os.path.splitext(os.path.basename(input))

        
        # get outline
        try:
            outlines = self.document.get_outlines()
        except Exception as e:
            logger.warning(f"loading PDF file: {e}")
            self.outline_flag = False
        
        
        if not self.outline_flag:

            with open(input, "rb") as file:
                for idx, page_layout in enumerate(extract_pages(file)):
                    content = ""
                    for element in page_layout:
                        if hasattr(element, "get_text"):
                            content = content + element.get_text()
                    chunk = Chunk(
                        id=Chunk.generate_hash_id(f"{basename}#{idx}"),
                        name=f"{basename}#{idx}",
                        content=content,
                    )
                    chunks.append(chunk)
            try:
                outline_chunks =  self.outline_chunk(chunks, basename)
            except Exception as e:
                raise RuntimeError(f"Error loading PDF file: {e}")
            if len(outline_chunks) > 0:
                chunks = outline_chunks
                
        else:
            split_words = []
        
            for item in outlines:
                level, title, dest, a, se = item
                split_words.append(title.strip().replace(" ",""))
            # save the outline position in content
            try:
                text = extract_text(input)

            except Exception as e:
                raise RuntimeError(f"Error loading PDF file: {e}")

            cleaned_pages = [
                self._process_single_page(x, "", False, False) for x in text
            ]
            sentences = []
            for cleaned_page in cleaned_pages:
                sentences += cleaned_page

            content = "".join(sentences)
            positions = [(input,0)]
            for split_word in split_words:
                pattern = re.compile(split_word)
                for i,match in enumerate(re.finditer(pattern, content)):
                    if i == 1:
                        start, end = match.span()
                        positions.append((split_word,start))
            
            for idx,position in enumerate(positions):
                chunk = Chunk(
                    id = Chunk.generate_hash_id(f"{basename}#{position[0]}"),
                    name=f"{basename}#{position[0]}",
                    content=content[position[1]:positions[idx+1][1] if idx+1 < len(positions) else None],
                )
                chunks.append(chunk)

        return chunks


if __name__ == '__main__':
    reader = PDFReader(split_using_outline=True)
    pdf_path = os.path.join(os.path.dirname(__file__),"../../../../tests/builder/data/aiwen.pdf")
    chunk = reader.invoke(pdf_path)
    print(chunk)


from kag.common.base.prompt_op import PromptOp
from kag.solver.logic.core_modules.common.base_model import LogicNode
from kag.solver.logic.core_modules.common.one_hop_graph import KgGraph
from kag.solver.logic.core_modules.common.schema_utils import SchemaUtils
from kag.solver.logic.core_modules.op_executor.op_executor import OpExecutor


class JudgementOp(OpExecutor):
    def __init__(self, nl_query: str, kg_graph: KgGraph, schema: SchemaUtils, debug_info: dict, **kwargs):
        super().__init__(nl_query, kg_graph, schema, debug_info, **kwargs)
        self.prompt = PromptOp.load(self.biz_scene, "deduce_judge")(
            language=self.language
        )

    def executor(self, logic_node: LogicNode, req_id: str, param: dict) -> list:
        history_qa_pair = self.debug_info.get("sub_qa_pair", [])
        qa_pair = "\n".join([f"Q: {q}\nA: {a}" for q, a in history_qa_pair])
        spo_info = self.kg_graph.to_evidence()
        information = str(spo_info) + "\n" + qa_pair
        if_answered, answer = self.llm_module.invoke({'instruction': self.nl_query, 'memory': information},
                                                     self.prompt, with_json_parse=False, with_except=True)
        return [if_answered, answer]


import json
import logging
import os
import re
import time
from typing import List

from kag.common.base.prompt_op import PromptOp
from kag.common.llm.client import LLMClient
from kag.solver.logic.core_modules.common.one_hop_graph import KgGraph, EntityData, OneHopGraphData, \
    RelationData
from kag.solver.logic.core_modules.common.schema_utils import SchemaUtils
from kag.solver.logic.core_modules.common.text_sim_by_vector import TextSimilarity
from kag.solver.logic.core_modules.parser.logic_node_parser import GetSPONode

logger = logging.getLogger()


def default_change_en_2_zh(s, p, o, schema: SchemaUtils):
    if schema is None:
        return p
    item = f"{s}_{p}_{o}"
    if item and item in schema.spo_en_zh.keys():
        return schema.get_spo_with_p(schema.spo_en_zh[item])
    item = f"{o}_{p}_{s}"
    if item and item in schema.spo_en_zh.keys():
        return schema.get_spo_with_p(schema.spo_en_zh[item])

    attr_en_zh = schema.get_attr_en_zh_by_label(s)
    if p in attr_en_zh.keys():
        return attr_en_zh[p]
    return p


def change_en_2_zh(s, p, o, schema: SchemaUtils):
    return f"{s}_{p}_{o}"


def split_value(value):
    value = value.strip()
    pattern = re.compile(r'[,、; ，]+')
    return pattern.split(value)


class RetrievalSpoBase:
    def __init__(self):
        pass

    def match_spo(self, n: GetSPONode, one_hop_graph_list: List[OneHopGraphData]):
        pass


class ExactMatchRetrievalSpo(RetrievalSpoBase):
    def __init__(self, schema):
        super().__init__()
        self.schema: SchemaUtils = schema

    def _prase_attribute_relation(self, one_graph, std_p: str, attr_value: str):
        # new a RelationData
        prop_entity = EntityData()
        prop_entity.biz_id = attr_value
        prop_entity.name = attr_value
        prop_entity.type = "attribute"
        prop_entity.type_zh = "文本"

        return self._prase_entity_relation(one_graph, std_p, prop_entity)

    def _prase_entity_relation(self, one_graph, std_p: str, o_value: EntityData):
        s_entity = one_graph.s
        o_entity = o_value
        if one_graph.s_alias_name == "o":
            o_entity = one_graph.s
            s_entity = o_value
        if o_value.description is None or o_value.description == '':
            o_value.description = f"{s_entity.name} {std_p} {o_entity.name}"
        return RelationData.from_prop_value(s_entity, std_p, o_entity)

    def _std_best_p_with_value_and_p_name(self, n: GetSPONode, one_graph: OneHopGraphData):
        """
        :param one_graph:
        :return: list(RelationData)
        """
        debug_info = {
            "el": [],
            "el_detail": [],
            "std_out": []
        }
        logger.debug(f"std_best_p_with_value_and_p_name begin std " + str(n))
        un_std_p_list = n.p.get_entity_type_or_zh_list()
        final_result_list = []
        if len(un_std_p_list) == 0:
            # return all
            result = []
            if len(one_graph.in_relations) > 0:
                for k in one_graph.in_relations.keys():
                    result = one_graph.in_relations[k] + result
            if len(one_graph.out_relations) > 0:
                for k in one_graph.out_relations.keys():
                    result = one_graph.out_relations[k] + result
            final_result_list = final_result_list + result

        for un_std_p in un_std_p_list:
            target_value = n.o.entity_name if one_graph.s_alias_name == "s" else n.s.entity_name
            target_node = n.o if one_graph.s_alias_name == "s" else n.s
            relation_name_set = one_graph.get_s_all_relation_name()
            attribute_name_set = one_graph.get_s_all_attribute_name()
            candi_name_set = relation_name_set + attribute_name_set

            def find_best_match_p_name(p: str, candi_set: list):
                if p in candi_set:
                    return p
                return None

            std_p = find_best_match_p_name(un_std_p, candi_name_set)
            debug_info['std_out'].append({
                "un_std_p": un_std_p,
                "candi_name_set": candi_name_set,
                "std_p": std_p if std_p is not None else ''
            })
            if std_p is None:
                continue

            get_data_from_rel = False
            if std_p in relation_name_set and std_p in attribute_name_set:
                if not target_node.is_attribute:
                    get_data_from_rel = True
            elif std_p in relation_name_set:
                get_data_from_rel = True

            if get_data_from_rel:
                relation_data = one_graph.get_std_relation_value(std_p)
            else:
                logger.info(f"relation with el: un std p is " + un_std_p + ", std p is " + std_p)
                value = one_graph.get_std_attribute_value(std_p)
                if value is None or value == "":
                    continue
                # new a RelationData
                relation_data = [self._prase_attribute_relation(one_graph, std_p, value)]
            if target_value is not None:
                for r in relation_data:
                    candi_target_value = r.end_entity.name if one_graph.s_alias_name == "s" else r.start_entity.name
                    if candi_target_value == target_value:
                        final_result_list.append(r)
                        continue
            else:
                final_result_list = final_result_list + relation_data
        return final_result_list, debug_info

    def match_spo(self, n: GetSPONode, one_hop_graph_list: List[OneHopGraphData]):
        matched_flag = False
        one_kg_graph = KgGraph()
        one_kg_graph.query_graph[n.p.alias_name] = {
            "s": n.s.alias_name,
            "p": n.p.alias_name,
            "o": n.o.alias_name
        }
        for tmp_one_hop_graph in one_hop_graph_list:
            rel_set, recall_debug_info = self._std_best_p_with_value_and_p_name(n, tmp_one_hop_graph)
            if len(rel_set) > 0:
                one_kg_graph_ = KgGraph()
                recall_alias_name = n.s.alias_name if tmp_one_hop_graph.s_alias_name == "s" else n.o.alias_name
                one_kg_graph_.entity_map[recall_alias_name] = [tmp_one_hop_graph.s]
                one_kg_graph_.edge_map[n.p.alias_name] = rel_set
                one_kg_graph.merge_kg_graph(one_kg_graph_)
        spo_set = one_kg_graph.get_entity_by_alias(n.p.alias_name)
        if spo_set is not None and len(spo_set) != 0:
            matched_flag = True
        return one_kg_graph, matched_flag


class FuzzyMatchRetrievalSpo(RetrievalSpoBase):
    def __init__(self, text_similarity: TextSimilarity = None, llm: LLMClient=None):
        super().__init__()
        model = eval(os.getenv("KAG_LLM"))
        self.llm: LLMClient = llm or LLMClient.from_config(model)
        self.text_similarity = text_similarity or TextSimilarity()
        self.cached_map = {}

        self.biz_scene = os.getenv("KAG_PROMPT_BIZ_SCENE", "default")
        self.language = os.getenv("KAG_PROMPT_LANGUAGE", "en")

    def get_unstd_p_text(self, n: GetSPONode):
        un_std_p = n.p.get_entity_first_type_or_zh()
        start_value_type = n.s.get_entity_first_type_or_zh()
        if start_value_type == "Others":
            start_value_type = "Entity"
        target_value_type = n.o.get_entity_first_type_or_zh()
        if target_value_type == "Others":
            target_value_type = "Entity"
        un_std_p = f"{start_value_type}{'[' + n.s.entity_name + ']' if n.s.entity_name is not None else ''} {un_std_p} {target_value_type}{'[' + n.o.entity_name + ']' if n.o.entity_name is not None else ''}"
        return un_std_p

    def _choosed_by_llm(self, question, mention, candis):
        resp_plan_prompt = PromptOp.load(self.biz_scene, "spo_retrieval")(
            language=self.language
        )
        return self.llm.invoke({
            'question': question,
            'mention': mention,
            'candis': candis
        }, resp_plan_prompt, with_json_parse=False, with_except=True)

    def select_relation(self, p_mention, p_candis, query='', topk=1, params={}):
        if not p_mention:
            print('p_mention is none')
            return None
        if p_mention in self.cached_map.keys():
            cached_set = self.cached_map[p_mention]
            intersection = list(set(cached_set) & set(p_candis))
        else:
            intersection = []
        if len(intersection) == 0:
            res = ''
            try:
                res = self._choosed_by_llm(query, p_mention, p_candis)
                res = res.replace("Output:", "output:")
                if "output:" in res:
                    res = re.search('output:(.*)', res).group(1).strip()
                if res != '':
                    res = json.loads(res.replace("'", '"'))
                    for res_ in res:
                        self.cached_map[p_mention] = self.cached_map.get(p_mention, []) + [res_]
                        intersection.append(res_)
            except:
                logger.warning(f"retrieval_spo json failed：query={query},  res={res}")
        return [[x, 1.0] for x in intersection]

    def find_best_match_p_name_by_model(self, query: str, p: str, candi_set: dict):
        if p in candi_set:
            return [p, candi_set[p]]
        spo_retrieved = []
        sen_condi_set = []
        spo_name_map = {}
        for p_name, spo_l in candi_set.items():
            if p_name.startswith("_") or p_name == "id" or p_name == 'content':
                continue
            for spo in spo_l:
                spo_name_map[spo] = p_name
            sen_condi_set += spo_l
        result = self.select_relation(p, sen_condi_set, query=query)
        logger.debug(f"retrieval_relation: p={p}, candi_set={sen_condi_set}, p_std result={result}")

        if result is None or len(result) == 0:
            return spo_retrieved

        for result_ in result:
            spo = result_[0]
            spo_p_name = spo_name_map.get(spo, None)
            spo_retrieved.append([spo, spo_p_name])
        return spo_retrieved

    def match_spo(self, n: GetSPONode, one_hop_graph_list: List[OneHopGraphData]):
        matched_flag = False
        one_kg_graph = KgGraph()
        # sort graph
        unstd_p_text = self.get_unstd_p_text(n)
        all_spo_text = []
        revert_value_p_map = {}
        revert_graph_map = {}
        for one_hop_graph in one_hop_graph_list:
            for k, v_set in one_hop_graph.get_s_all_relation_spo().items():
                for v in v_set:
                    all_spo_text.append(v)
                    revert_value_p_map[v] = k
                    revert_graph_map[v] = one_hop_graph
            for k, v_set in one_hop_graph.get_s_all_attribute_spo().items():
                for v in v_set:
                    all_spo_text.append(v)
                    revert_value_p_map[v] = k
                    revert_graph_map[v] = one_hop_graph
        start_time = time.time()
        tok5_res = self.text_similarity.text_sim_result(n.sub_query, all_spo_text, 5, low_score=0.3)
        logger.debug(f" _get_spo_value_in_one_hop_graph_set text similarity cost={time.time() - start_time}")

        if len(tok5_res) == 0:
            return one_kg_graph, matched_flag

        matched_flag = True

        candi_name_set = {}
        for res in tok5_res:
            k = revert_value_p_map[res[0]]
            if k in candi_name_set.keys():
                candi_name_set[k].append(res[0])
            else:
                candi_name_set[k] = [res[0]]
        start_time = time.time()
        spo_retrieved = self.find_best_match_p_name_by_model(n.sub_query, unstd_p_text,
                                                             candi_name_set)
        logger.debug(
            f"_get_spo_value_in_one_hop_graph_set find_best_match_p_name_by_entity_list cost={time.time() - start_time}")
        total_one_kg_graph = KgGraph()
        total_one_kg_graph.query_graph[n.p.alias_name] = {
            "s": n.s.alias_name,
            "p": n.p.alias_name,
            "o": n.o.alias_name
        }
        for std_spo_text, std_p in spo_retrieved:
            if std_p is None or std_p == '':
                continue
            one_hop_graph = revert_graph_map[std_spo_text]
            rel_set = one_hop_graph.get_std_p_value_by_spo_text(std_p, std_spo_text)
            one_kg_graph_ = KgGraph()
            recall_alias_name = n.s.alias_name if one_hop_graph.s_alias_name == "s" else n.o.alias_name
            one_kg_graph_.entity_map[recall_alias_name] = [one_hop_graph.s]
            one_kg_graph_.edge_map[n.p.alias_name] = rel_set
            total_one_kg_graph.merge_kg_graph(one_kg_graph_)
        return total_one_kg_graph, matched_flag




from typing import List

from .evaUtils import get_em_f1


class Evaluate():

    """
    provide evaluation for benchmarks, such as em、f1、answer_similarity, answer_correctness
    """
    def __init__(self, embedding_factory = "text-embedding-ada-002"):
        self.embedding_factory = embedding_factory

    def evaForSimilarity(self, predictionlist: List[str], goldlist: List[str]):
        """
        evaluate the similarity between prediction and gold #TODO
        """
        # data_samples = {  
        #     'question': [],
        #     'answer': predictionlist,
        #     'ground_truth': goldlist
        # }
        # dataset = Dataset.from_dict(data_samples)
        # run_config = RunConfig(timeout=240, thread_timeout=240, max_workers=16)
        # embeddings = embedding_factory(self.embedding_factory, run_config)
        #
        # score = evaluate(dataset, metrics=[answer_similarity], embeddings = embeddings, run_config=run_config)
        # return np.average(score.to_pandas()[['answer_similarity']])
        return 0.0


    def getBenchMark(self, predictionlist: List[str], goldlist: List[str]):
        """
        Calculates and returns evaluation metrics between predictions and ground truths.

        This function evaluates the match between predictions and ground truths by calculating
        the exact match (EM) and F1 score, as well as answer similarity.

        Parameters:
        predictionlist (List[str]): List of predicted values from the model.
        goldlist (List[str]): List of actual ground truth values.

        Returns:
        dict: Dictionary containing EM, F1 score, and answer similarity.
        """
        # Initialize total metrics
        total_metrics = {'em': 0.0, 'f1': 0.0, 'answer_similarity': 0.0}
        
        # Iterate over prediction and gold lists to calculate EM and F1 scores
        for prediction, gold in zip(predictionlist, goldlist):
            em, f1 = get_em_f1(prediction, gold)  # Call external function to calculate EM and F1
            total_metrics['em'] += em  # Accumulate EM score
            total_metrics['f1'] += f1  # Accumulate F1 score
        
        # Calculate average EM and F1 scores
        total_metrics['em'] /= len(predictionlist)
        total_metrics['f1'] /= len(predictionlist)
        
        # Call method to calculate answer similarity
        total_metrics['answer_similarity'] = self.evaForSimilarity(predictionlist, goldlist)

        # Return evaluation metrics dictionary
        return total_metrics




# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os
import ast
import re
import json
import time
import uuid
import html
from binascii import b2a_hex
from datetime import datetime
from pathlib import Path
from typing import Union, Dict, List, Any
from urllib import request
from collections import defaultdict

from openai import OpenAI
import logging

import requests
import traceback
from Crypto.Cipher import AES
from requests import RequestException

from kag.common import arks_pb2
from kag.common.base.prompt_op import PromptOp
from kag.common.llm.config import VLLMConfig

from kag.common.llm.client.llm_client import LLMClient


# logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

class VLLMClient(LLMClient):
    def __init__(self, llm_config: VLLMConfig):
        self.model = llm_config.model
        self.base_url = llm_config.base_url
        self.param = {}

    def sync_request(self, prompt):
        # import pdb; pdb.set_trace()
        self.param["messages"] = prompt
        self.param["model"] = self.model

        response = requests.post(
            self.base_url,
            data=json.dumps(self.param),
            headers={"Content-Type": "application/json"},
        )

        data = response.json()
        content = data["choices"][0]["message"]["content"]
        content = content.replace("&rdquo;", "”").replace("&ldquo;", "“")
        content = content.replace("&middot;", "")
        return content

    def __call__(self, prompt):
        content = [
          {"role": "user", "content": prompt}
          ]
        return self.sync_request(content)

    def call_with_json_parse(self, prompt):
        content = [{"role": "user", "content": prompt}]
        rsp = self.sync_request(content)
        _end = rsp.rfind("```")
        _start = rsp.find("```json")
        if _end != -1 and _start != -1:
            json_str = rsp[_start + len("```json"): _end].strip()
        else:
            json_str = rsp
        try:
            json_result = json.loads(json_str)
        except:
            return rsp
        return json_result



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import os
from collections import defaultdict
from typing import List

from kag.builder.model.sub_graph import SubGraph
from knext.common.base.runnable import Input, Output
from kag.common.vectorizer import Vectorizer
from kag.interface.builder.vectorizer_abc import VectorizerABC
from knext.schema.client import SchemaClient
from knext.project.client import ProjectClient
from knext.schema.model.base import IndexTypeEnum


class EmbeddingVectorPlaceholder(object):
    def __init__(self, number, properties, vector_field, property_key, property_value):
        self._number = number
        self._properties = properties
        self._vector_field = vector_field
        self._property_key = property_key
        self._property_value = property_value
        self._embedding_vector = None

    def replace(self):
        if self._embedding_vector is not None:
            self._properties[self._vector_field] = self._embedding_vector

    def __repr__(self):
        return repr(self._number)


class EmbeddingVectorManager(object):
    def __init__(self):
        self._placeholders = []

    def _create_vector_field_name(self, property_key):
        from kag.common.utils import to_snake_case

        name = f"{property_key}_vector"
        name = to_snake_case(name)
        return "_" + name

    def get_placeholder(self, properties, vector_field):
        for property_key, property_value in properties.items():
            field_name = self._create_vector_field_name(property_key)
            if field_name != vector_field:
                continue
            if not property_value:
                return None
            if not isinstance(property_value, str):
                message = f"property {property_key!r} must be string to generate embedding vector"
                raise RuntimeError(message)
            num = len(self._placeholders)
            placeholder = EmbeddingVectorPlaceholder(
                num, properties, vector_field, property_key, property_value
            )
            self._placeholders.append(placeholder)
            return placeholder
        return None

    def _get_text_batch(self):
        text_batch = dict()
        for placeholder in self._placeholders:
            property_value = placeholder._property_value
            if property_value not in text_batch:
                text_batch[property_value] = list()
            text_batch[property_value].append(placeholder)
        return text_batch

    def _generate_vectors(self, vectorizer, text_batch):
        texts = list(text_batch)
        if not texts:
            return []
        vectors = vectorizer.vectorize(texts)
        return vectors

    def _fill_vectors(self, vectors, text_batch):
        for vector, (_text, placeholders) in zip(vectors, text_batch.items()):
            for placeholder in placeholders:
                placeholder._embedding_vector = vector

    def batch_generate(self, vectorizer):
        text_batch = self._get_text_batch()
        vectors = self._generate_vectors(vectorizer, text_batch)
        self._fill_vectors(vectors, text_batch)

    def patch(self):
        for placeholder in self._placeholders:
            placeholder.replace()


class EmbeddingVectorGenerator(object):
    def __init__(self, vectorizer, vector_index_meta=None, extra_labels=("Entity",)):
        self._vectorizer = vectorizer
        self._extra_labels = extra_labels
        self._vector_index_meta = vector_index_meta or {}

    def batch_generate(self, node_batch):
        manager = EmbeddingVectorManager()
        vector_index_meta = self._vector_index_meta
        for node_item in node_batch:
            label, properties = node_item
            labels = [label]
            if self._extra_labels:
                labels.extend(self._extra_labels)
            for label in labels:
                if label not in vector_index_meta:
                    continue
                for vector_field in vector_index_meta[label]:
                    if vector_field in properties:
                        continue
                    placeholder = manager.get_placeholder(properties, vector_field)
                    if placeholder is not None:
                        properties[vector_field] = placeholder
        manager.batch_generate(self._vectorizer)
        manager.patch()


class BatchVectorizer(VectorizerABC):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.project_id = self.project_id or os.getenv("KAG_PROJECT_ID")
        self._init_graph_store()
        self.vec_meta = self._init_vec_meta()
        self.vectorizer = Vectorizer.from_config(self.vectorizer_config)

    def _init_graph_store(self):
        """
        Initializes the Graph Store client.

        This method retrieves the graph store configuration from environment variables and the project ID.
        It then fetches the project configuration using the project ID and updates the graph store configuration
        with any additional settings from the project. Finally, it creates and initializes the graph store client
        using the updated configuration.

        Args:
            project_id (str): The id of project.

        Returns:
            GraphStore
        """
        graph_store_config = eval(os.getenv("KAG_GRAPH_STORE", "{}"))
        vectorizer_config = eval(os.getenv("KAG_VECTORIZER", "{}"))
        config = ProjectClient().get_config(self.project_id)
        graph_store_config.update(config.get("graph_store", {}))
        vectorizer_config.update(config.get("vectorizer", {}))
        self.vectorizer_config = vectorizer_config

    def _init_vec_meta(self):
        vec_meta = defaultdict(list)
        schema_client = SchemaClient(project_id=self.project_id)
        spg_types = schema_client.load()
        for type_name, spg_type in spg_types.items():
            for prop_name, prop in spg_type.properties.items():
                if prop_name == "name" or prop.index_type in [IndexTypeEnum.Vector, IndexTypeEnum.TextAndVector]:
                    vec_meta[type_name].append(self._create_vector_field_name(prop_name))
        return vec_meta

    def _create_vector_field_name(self, property_key):
        from kag.common.utils import to_snake_case

        name = f"{property_key}_vector"
        name = to_snake_case(name)
        return "_" + name

    def _generate_embedding_vectors(self, vectorizer: Vectorizer, input: SubGraph) -> SubGraph:
        node_list = []
        node_batch = []
        for node in input.nodes:
            if not node.id or not node.name:
                continue
            properties = {"id": node.id, "name": node.name}
            properties.update(node.properties)
            node_list.append((node, properties))
            node_batch.append((node.label, properties.copy()))
        generator = EmbeddingVectorGenerator(vectorizer, self.vec_meta)
        generator.batch_generate(node_batch)
        for (node, properties), (_node_label, new_properties) in zip(
            node_list, node_batch
        ):
            for key, value in properties.items():
                if key in new_properties and new_properties[key] == value:
                    del new_properties[key]
            node.properties.update(new_properties)
        return input

    def invoke(self, input: Input, **kwargs) -> List[Output]:
        modified_input = self._generate_embedding_vectors(self.vectorizer, input)
        return [modified_input]



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os
import json
from pathlib import Path
from typing import Union, Dict, List, Any
import logging
import traceback
import yaml

from kag.common.base.prompt_op import PromptOp
from kag.common.llm.config import *


logger = logging.getLogger(__name__)


class LLMClient:
    # Define the model type
    model: str

    config_cls_map = {
        "maas": OpenAIConfig,
        "vllm": VLLMConfig,
        "ollama": OllamaConfig,
    }

    def __init__(self, **kwargs):
        self.model = kwargs.get("model", None)
    
    @classmethod
    def get_config_cls(self,config:dict):
        client_type = config.get("client_type", None)
        return LLMClient.config_cls_map.get(client_type, None)
        
    @classmethod
    def get_llm_cls(self,config: LLMConfig):
        from kag.common.llm.client import VLLMClient,OpenAIClient,OllamaClient
        return {
            VLLMConfig: VLLMClient,
            OpenAIConfig: OpenAIClient,
            OllamaConfig: OllamaClient,
        }[config.__class__]

    @classmethod
    def from_config(cls, config: Union[str, dict]):
        """
        Initialize an LLMClient instance from a configuration file or dictionary.

        :param config: Path to a configuration file or a configuration dictionary
        :return: Initialized LLMClient instance
        :raises FileNotFoundError: If the configuration file is not found
        :raises ValueError: If the model type is unsupported
        """
        if isinstance(config, str):
            config_path = Path(config)
            if config_path.is_file():
                try:
                    with open(config_path, "r") as f:
                        nn_config = yaml.safe_load(f)
                except:
                    logger.error(f"Failed to parse config file")
                    raise
            else:
                logger.error(f"Config file not found: {config}")
                raise FileNotFoundError(f"Config file not found: {config}")
        else:
            # If config is already a dictionary, use it directly
            nn_config = config
        
        config_cls = LLMClient.get_config_cls(nn_config)
        if config_cls is None:
            logger.error(f"Unsupported model type: {nn_config.get('client_type', None)}")
            raise ValueError(f"Unsupported model type")
        llm_config = config_cls(**nn_config)
        llm_cls = LLMClient.get_llm_cls(llm_config)
        return llm_cls(llm_config)
        
            
    def __call__(self, prompt: Union[str, dict, list]) -> str:
        """
        Perform inference on the given prompt and return the result.

        :param prompt: Input prompt for inference
        :return: Inference result
        :raises NotImplementedError: If the subclass has not implemented this method
        """
        raise NotImplementedError

    def call_with_json_parse(self, prompt: Union[str, dict, list]):
        """
        Perform inference on the given prompt and attempt to parse the result as JSON.

        :param prompt: Input prompt for inference
        :return: Parsed result
        :raises NotImplementedError: If the subclass has not implemented this method
        """
        res = self(prompt)
        _end = res.rfind("```")
        _start = res.find("```json")
        if _end != -1 and _start != -1:
            json_str = res[_start + len("```json"): _end].strip()
        else:
            json_str = res
        try:
            json_result = json.loads(json_str)
        except:
            return res
        return json_result

    def invoke(self, variables: Dict[str, Any], prompt_op: PromptOp, with_json_parse: bool = True,
               with_except: bool = False):
        """
        Call the model and process the result.

        :param variables: Variables used to build the prompt
        :param prompt_op: Prompt operation object for building and parsing prompts
        :param with_json_parse: Whether to attempt parsing the response as JSON
        :param with_except: Whether to raise exception
        :return: Processed result list
        """
        result = []
        prompt = prompt_op.build_prompt(variables)
        logger.debug(f"Prompt: {prompt}")
        if not prompt:
            return result
        response = ""
        try:
            response = self.call_with_json_parse(prompt=prompt) if with_json_parse else self(prompt)
            logger.debug(f"Response: {response}")
            result = prompt_op.parse_response(response, model=self.model, **variables)
            logger.debug(f"Result: {result}")
        except Exception as e:
            import traceback
            logger.debug(f"Error {e} during invocation: {traceback.format_exc()}")
            if with_except:
                raise RuntimeError(f"call llm exception! llm output = {response} , llm input={prompt}, err={e}")
        return result

    def batch(self, variables: Dict[str, Any], prompt_op: PromptOp, with_json_parse: bool = True) -> List:
        """
        Batch process prompts.

        :param variables: Variables used to build the prompts
        :param prompt_op: Prompt operation object for building and parsing prompts
        :param with_json_parse: Whether to attempt parsing the response as JSON
        :return: List of all processed results
        """
        results = []
        prompts = prompt_op.build_prompt(variables)
        # If there is only one prompt, call the `invoke` method directly
        if isinstance(prompts, str):
            return self.invoke(variables, prompt_op, with_json_parse=with_json_parse)

        for idx, prompt in enumerate(prompts, start=0):
            logger.debug(f"Prompt_{idx}: {prompt}")
            try:
                response = self.call_with_json_parse(prompt=prompt) if with_json_parse else self(prompt)
                logger.debug(f"Response_{idx}: {response}")
                result = prompt_op.parse_response(response, idx=idx, model=self.model, **variables)
                logger.debug(f"Result_{idx}: {result}")
                results.extend(result)
            except Exception as e:
                logger.error(f"Error processing prompt {idx}: {e}")
                logger.debug(traceback.format_exc())
                continue
        return results
    


# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import os
from typing import List, Type,Union

from docx import Document

from kag.builder.component.reader import MarkDownReader
from kag.builder.model.chunk import Chunk
from kag.interface.builder import SourceReaderABC
from knext.common.base.runnable import Input, Output

from kag.common.llm.client import LLMClient
from kag.builder.prompt.outline_prompt import OutlinePrompt

def split_txt(content):
    from modelscope.outputs import OutputKeys
    from modelscope.pipelines import pipeline
    from modelscope.utils.constant import Tasks

    p = pipeline(
        task=Tasks.document_segmentation,
        model='damo/nlp_bert_document-segmentation_chinese-base')

    result = p(documents=content)
    result = result[OutputKeys.TEXT]
    
    res = [r for r in result.split('\n\t') if len(r) > 0]
    
    return res


    
class DocxReader(SourceReaderABC):
    """
    A class for reading Docx files, inheriting from SourceReader.
    This class is specifically designed to extract text content from Docx files and generate Chunk objects based on the extracted content.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.split_level = kwargs.get("split_level", 3)
        self.split_using_outline = kwargs.get("split_using_outline", True)
        self.outline_flag = True
        self.llm = self._init_llm()
        language = os.getenv("KAG_PROMPT_LANGUAGE", "zh")
        self.prompt = OutlinePrompt(language)

    @property
    def input_types(self) -> Type[Input]:
        return str

    @property
    def output_types(self) -> Type[Output]:
        return Chunk
    
    def outline_chunk(self, chunk: Union[Chunk, List[Chunk]],basename) -> List[Chunk]:
        if isinstance(chunk, Chunk):
            chunk = [chunk]
        outlines = []
        for c in chunk:
            outline = self.llm.invoke({"input": c.content}, self.prompt)
            outlines.extend(outline)
        content = "\n".join([c.content for c in chunk])
        chunks = self.sep_by_outline(content, outlines,basename)
        return chunks
    
    def sep_by_outline(self,content,outlines,basename):
        position_check = []
        for outline in outlines:
            start = content.find(outline)
            position_check.append((outline,start))
        chunks = []
        for idx,pc in enumerate(position_check):
            chunk = Chunk(
                id = Chunk.generate_hash_id(f"{basename}#{pc[0]}"),
                name=f"{basename}#{pc[0]}",
                content=content[pc[1]:position_check[idx+1][1] if idx+1 < len(position_check) else len(position_check)],
            )
            chunks.append(chunk)
        return chunks

    @staticmethod
    def _extract_text_from_docx(doc: Document) -> str:
        """
        Extracts text from a Docx document.

        This method iterates through all paragraphs in the provided Docx document,
        appending each paragraph's text to a list, and then joins these texts into
        a single string separated by newline characters, effectively extracting the
        entire text content of the document.

        Args:
            doc (Document): A Document object representing the Docx file from which
                            text is to be extracted.

        Returns:
            str: A string containing all the text from the Docx document, with paragraphs
                 separated by newline characters.
        """
        full_text = []
        for para in doc.paragraphs:
            full_text.append(para.text)
        return full_text
    
    def _get_title_from_text(self, text: str) -> str:
        text = text.strip()
        title = text.split('\n')[0]
        text = "\n".join(text.split('\n'))
        return title,text

    def invoke(self, input: Input, **kwargs) -> List[Output]:
        """
        Processes the input Docx file, extracts its text content, and generates a Chunk object.

        Args:
            input (Input): The file path of the Docx file to be processed.
            **kwargs: Additional keyword arguments, not used in the current implementation.

        Returns:
            List[Output]: A list containing a single Chunk object with the extracted text.

        Raises:
            ValueError: If the input is empty.
            IOError: If the file cannot be read or the text extraction fails.
        """

        if not input:
            raise ValueError("Input cannot be empty")
        
        chunks = []
        
        try:
            doc = Document(input)
            full_text = self._extract_text_from_docx(doc)
            content = "\n".join(full_text)
        except OSError as e:
            raise IOError(f"Failed to read file: {input}") from e

        basename, _ = os.path.splitext(os.path.basename(input))

        for text in full_text:
            title,text = self._get_title_from_text(text)
            chunk = Chunk(
                id=Chunk.generate_hash_id(f"{basename}#{title}"),
                name=f"{basename}#{title}",
                content=text,
            )
            chunks.append(chunk)

        if len(chunks) < 2:
            chunks = self.outline_chunk(chunks,basename)
        
        if len(chunks) < 2:
            semantic_res = split_txt(content)
            chunks = [Chunk(
                id=Chunk.generate_hash_id(input+"#"+r[:10]),
                name=basename+"#"+r[:10],
                content=r,
            ) for r in semantic_res]

        return chunks


if __name__== "__main__":
    reader = DocxReader()
    print(reader.output_types)
    file_path = os.path.dirname(__file__)
    res = reader.invoke(os.path.join(file_path,"../../../../tests/builder/data/test_docx.docx"))
    print(res)


from abc import ABC, abstractmethod

from kag.solver.common.base import KagBaseModule


class KagMemoryABC(KagBaseModule, ABC):
    def __init__(self, **kwargs):
        """
        Initializes the KagMemory module with an LLM module.

        :param llm_module: The LLM module to use.
        """
        super().__init__(**kwargs)

    @abstractmethod
    def save_memory(self, solved_answer, supporting_fact, instruction):
        """
        Saves the solved answer, supporting facts, and instruction.

        :param solved_answer: The solved answer.
        :param supporting_fact: The supporting fact.
        :param instruction: The instruction.
        """
        pass

    @abstractmethod
    def get_solved_answer(self) -> str:
        """
        Retrieves the solved answer.

        :return: The solved answer.
        """

    @abstractmethod
    def serialize_memory(self) -> str:
        """
        Serializes the memory to str.

        :return: Serialized memory data with str format.
        """

    @abstractmethod
    def refresh(self):
        """
        Refreshes the memory.

        This method is used to reset the memory state.
        """


# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import logging
import os
import sys
from configparser import ConfigParser as CP
from pathlib import Path
from typing import Union, Optional

import kag.common as common

class ConfigParser(CP):
    def __init__(self,defaults=None):
        CP.__init__(self,defaults=defaults)
    def optionxform(self, optionstr):
        return optionstr


LOCAL_SCHEMA_URL = "http://localhost:8887"
DEFAULT_KAG_CONFIG_FILE_NAME = "default_config.cfg"
DEFAULT_KAG_CONFIG_PATH = os.path.join(common.__path__[0], DEFAULT_KAG_CONFIG_FILE_NAME)
KAG_CFG_PREFIX = "KAG"


def init_env():
    """Initialize environment to use command-line tool from inside a project
    dir. This sets the Scrapy settings module and modifies the Python path to
    be able to locate the project module.
    """
    project_cfg, root_path = get_config()

    init_kag_config(Path(root_path) / "kag_config.cfg")


def get_config():
    """
    Get kag config file as a ConfigParser.
    """
    local_cfg_path = _closest_cfg()
    local_cfg = ConfigParser()
    local_cfg.read(local_cfg_path)

    projdir = ""
    if local_cfg_path:
        projdir = str(Path(local_cfg_path).parent)
        if projdir not in sys.path:
            sys.path.append(projdir)

    return local_cfg, projdir


def _closest_cfg(
    path: Union[str, os.PathLike] = ".",
    prev_path: Optional[Union[str, os.PathLike]] = None,
) -> str:
    """
    Return the path to the closest .kag.cfg file by traversing the current
    directory and its parents
    """
    if prev_path is not None and str(path) == str(prev_path):
        return ""
    path = Path(path).resolve()
    cfg_file = path / "kag_config.cfg"
    if cfg_file.exists():
        return str(cfg_file)
    return _closest_cfg(path.parent, path)


def get_cfg_files():
    """
    Get global and local kag config files and paths.
    """
    local_cfg_path = _closest_cfg()
    local_cfg = ConfigParser()
    local_cfg.read(local_cfg_path)

    if local_cfg_path:
        projdir = str(Path(local_cfg_path).parent)
        if projdir not in sys.path:
            sys.path.append(projdir)

    return local_cfg, local_cfg_path



def init_kag_config(config_path: Union[str, Path] = None):
    if not config_path or isinstance(config_path, Path) and not config_path.exists():
        config_path = DEFAULT_KAG_CONFIG_PATH
    kag_cfg = ConfigParser()
    kag_cfg.read(config_path)
    os.environ["KAG_PROJECT_ROOT_PATH"] = os.path.abspath(os.path.dirname(config_path))

    for section in kag_cfg.sections():
        sec_cfg = {}
        for key, value in kag_cfg.items(section):
            item_cfg_key = f"{KAG_CFG_PREFIX}_{section}_{key}".upper()
            os.environ[item_cfg_key] = value
            sec_cfg[key] = value
        sec_cfg_key = f"{KAG_CFG_PREFIX}_{section}".upper()
        os.environ[sec_cfg_key] = str(sec_cfg)
        if section == "log":
            for key, value in kag_cfg.items(section):
                if key == "level":
                    logging.basicConfig(level=logging.getLevelName(value))
                    # neo4j log level set to be default error
                    logging.getLogger("neo4j.notifications").setLevel(logging.ERROR)
                    logging.getLogger("neo4j.io").setLevel(logging.INFO)
                    logging.getLogger("neo4j.pool").setLevel(logging.INFO)



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import json
import os
from typing import List, Type

from kag.builder.model.chunk import Chunk
from kag.interface.builder import SourceReaderABC
from knext.common.base.runnable import Input, Output


class HotpotqaCorpusReader(SourceReaderABC):
    @property
    def input_types(self) -> Type[Input]:
        """The type of input this Runnable object accepts specified as a type annotation."""
        return str

    @property
    def output_types(self) -> Type[Output]:
        """The type of output this Runnable object produces specified as a type annotation."""
        return Chunk

    def invoke(self, input: str, **kwargs) -> List[Output]:
        if os.path.exists(str(input)):
            with open(input, "r") as f:
                corpus = json.load(f)
        else:
            corpus = json.loads(input)
        chunks = []

        for item_key, item_value in corpus.items():
            chunk = Chunk(
                id=item_key,
                name=item_key,
                content="\n".join(item_value),
            )
            chunks.append(chunk)
        return chunks


class MusiqueCorpusReader(SourceReaderABC):
    @property
    def input_types(self) -> Type[Input]:
        """The type of input this Runnable object accepts specified as a type annotation."""
        return str

    @property
    def output_types(self) -> Type[Output]:
        """The type of output this Runnable object produces specified as a type annotation."""
        return Chunk

    def get_basename(self, file_name: str):
        base, ext = os.path.splitext(os.path.basename(file_name))
        return base

    def invoke(self, input: str, **kwargs) -> List[Output]:
        id_column = kwargs.get("id_column", "title")
        name_column = kwargs.get("name_column", "title")
        content_column = kwargs.get("content_column", "text")

        if os.path.exists(str(input)):
            with open(input, "r") as f:
                corpusList = json.load(f)
        else:
            corpusList = input
        chunks = []

        for item in corpusList:
            chunk = Chunk(
                id=item[id_column],
                name=item[name_column],
                content=item[content_column],
            )
            chunks.append(chunk)
        return chunks


class TwowikiCorpusReader(MusiqueCorpusReader):
    @property
    def input_types(self) -> Type[Input]:
        """The type of input this Runnable object accepts specified as a type annotation."""
        return str

    @property
    def output_types(self) -> Type[Output]:
        """The type of output this Runnable object produces specified as a type annotation."""
        return Chunk
